
from __future__ import annotations
from typing import Mapping, Any, Sequence

from dataclasses import dataclass
from datetime import datetime, timezone

from .datamemento import DatamementoPropertiesMixin, DatamementoDataclassesMixin


class WikiPageRevisionAuthorUser(DatamementoPropertiesMixin):
    """
    This class is pretty much the same as :class:`~.models.User` except
    the `awardee_karma`, `awarder_karma`, `total_karma` fields are missing.
    """
    class Subreddit:
        def __init__(self, d: Mapping[str, Any]) -> None:
            d = d['subreddit']
            self.name: str = d['display_name']
            ("")
            self.openness: str = d['subreddit_type']
            ("")
            self.subscriber_count: int = d['subscribers']
            ("")
            self.title: str = d['title']
            ("")
            self.public_description: str = d['public_description']
            ("")
            self.nsfw: bool = d['over_18']
            ("")

    class Me:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.is_friend: bool = d['is_friend']
            ("")
            self.is_blocked: bool = d['is_blocked']
            ("")

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.id36: str = d['id']
        ("")
        self.id: int = int(self.id36, 36)
        ("")
        self.name: str = d['name']
        ("")
        self.created_ut: int = int(d['created_utc'])
        ("")
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("")

        self.post_karma: int = d['link_karma']
        ("")
        self.comment_karma: int = d['comment_karma']
        ("")

        self.has_premium: bool = d['is_gold']
        ("")
        self.has_verified_email: bool = d['has_verified_email']
        ("")

        self.is_admin: bool = d['is_employee']
        ("")
        self.is_a_subreddit_moderator: bool = d['is_mod']
        ("")

        self.icon_img: str = d['icon_img']
        ("")

        self.subreddit: WikiPageRevisionAuthorUser.Subreddit = self.Subreddit(d)
        ("")
        self.me: WikiPageRevisionAuthorUser.Me = self.Me(d)
        ("")


@dataclass(repr=False, eq=False, frozen=True)
class WikiPage(DatamementoDataclassesMixin):
    d: Mapping[str, Any]
    body: str
    ("""
        The wiki page markdown content.
        """)
    body_html: str
    ("""
        The wiki page content as HTML.
        """)
    can_revise: bool
    ("""
        True if the current user may edit the wiki page.
        """)
    revision_uuid: str
    ("""
        The current revision UUID.
        """)
    revision_unixtime: int
    ("""
        UNIX timestamp of when the current revision was commited.
        """)
    revision_author: WikiPageRevisionAuthorUser
    ("""
        Author of the revision.
        """)
    revision_message: str
    ("""
        The current revision message.

        Up to 256 characters long.
        """)

@dataclass(repr=False, eq=False, frozen=True)
class WikiPageRevision(DatamementoDataclassesMixin):
    d: Mapping[str, Any]
    uuid: str
    ("""
        Revision UUID.
        """)
    unixtime: int
    ("""
        UNIX timestamp of when the current revision was commited.
        """)
    author: WikiPageRevisionAuthorUser
    ("""
        Author of the revision.
        """)
    message: str
    ("""
        Revision message.

        Up to 256 characters long.
        """)
    hidden: bool
    ("""
        True if the revision is hidden.
        """)

@dataclass(repr=False, eq=False, frozen=True)
class WikiPageSettings:
    permlevel: int
    ("""
        Permission level indicating who can edit this wiki page.

        "who can edit this page?"

        `0`: "use subreddit wiki permissions"
        `1`: "only approved wiki contributors for this page may edit"
        `2`: "only mods may edit and view"
        """)
    editors: Sequence[WikiPageRevisionAuthorUser]
    ("""
        List of users allowed to edit this page.
        """)
    indexed: bool
    ("""
        Whether this wiki page is indexed on the wiki page list.
        """)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from .my_account import MyAccount as BaseMyAccount

class MyAccount(BaseMyAccount):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")



"""Convert integers to strings representing numbers of various bases."""

from __future__ import annotations
from collections.abc import Iterable

def base_digits(base: int, n: int) -> Iterable[int]:
    """Yield the base digits of number `n`.

    A `ValueError` is raised if `base < 2` or `n <= 0`.

    ::

       >>> list(base_digits(10, 123))
       [3, 2, 1]
       >>> list(base_digits(16, 123))
       [11, 7]
       >>> list(base_digits(8, 123))
       [3, 7, 1]
       >>> list(base_digits(2, 123))
       [1, 1, 0, 1, 1, 1, 1]
    """
    if base < 2:
        raise ValueError("`base` must be at least 2")
    if n <= 0:
        raise ValueError("`n` must be positive")

    def g(base: int, n: int) -> Iterable[int]:
        while n:
            n, r = divmod(n, base)
            yield r

    return g(base, n)

def to_base(base: int, n: int, alphabet: str = "0123456789abcdefghijklmnopqrstuvwxyz") -> str:
    """Convert number `n` to base `base`."""
    if base > len(alphabet):
        raise ValueError("`alphabet` is not large enough for the given `base`")
    if n == 0:
        return alphabet[0]
    sign = '-' if n < 0 else ''
    numbers = base_digits(base, abs(n))
    symbols = [alphabet[i] for i in numbers]
    return sign + ''.join(reversed(symbols))

def to_base36(n: int, alphabet: str = "0123456789abcdefghijklmnopqrstuvwxyz") -> str:
    """Convert number `n` to base 36."""
    return to_base(36, n, alphabet=alphabet)




from __future__ import annotations
from typing import TypeVar, Sequence, Generic, Iterator, Optional

T = TypeVar('T')

# Abstract classes

class Paginator(Generic[T]):
    def __init__(self, *, limit: Optional[int] = None) -> None:
        self.limit: Optional[int] = limit
        ("")

    def __iter__(self) -> Iterator[Sequence[T]]:
        while page := self.fetch():
            yield page

    def fetch(self) -> Sequence[T]:
        raise NotImplementedError

class OffsetPaginator(Paginator[T]):
    def __init__(self, *, limit: Optional[int] = None, offset: Optional[int] = None) -> None:
        super().__init__(limit=limit)
        self.offset: Optional[int] = offset
        ("")

class CursorPaginator(Paginator[T]):
    def get_cursor(self) -> str:
        raise NotImplementedError

    def set_cursor(self, value: str) -> None:
        raise NotImplementedError


# Interfaces

class Bidirectional:
    direction: bool

class Resettable:
    def reset(self) -> None:
        raise NotImplementedError

class HasMore:
    def has_more(self) -> bool:
        raise NotImplementedError

    def set_has_more(self, value: bool) -> None:
        raise NotImplementedError


# Mixins

class HasMorePaginator(HasMore, Paginator[T]):
    def __iter__(self) -> Iterator[Sequence[T]]:
        if page := self.fetch():
            yield page
            while self.has_more() and (page := self.fetch()):
                yield page




from __future__ import annotations
from typing import Mapping, Any, Sequence, TypeVar, overload, Iterator, Union, final, Callable

from datetime import datetime, timezone

from .datamemento import DatamementoPropertiesMixin
from ..model_loaders.submission import load_submission
from .submission import Submission

class SubmissionCollectionInfo(DatamementoPropertiesMixin):
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.uuid: str = d['collection_id']
        ("")

        full_id36: str = d['subreddit_id']
        _, _, id36 = full_id36.partition('_')
        self.subreddit_id36: str = id36
        ("")
        self.subreddit_id: int = int(id36, 36)
        ("")

        full_id36 = d['author_id']
        _, _, id36 = full_id36.partition('_')
        self.author_id36: str = id36
        ("")
        self.author_id: int = int(id36, 36)
        ("")

        self.author_display_name: str = d['author_name']
        ("")
        self.title: str = d['title']
        ("")
        self.description: str = d['description']
        ("")

        self.created_ts: float = d['created_at_utc']
        ("")
        self.created_at: datetime = datetime.fromtimestamp(self.created_ts, timezone.utc)
        ("")
        self.last_post_added_ut: float = d['last_update_utc']
        ("")
        self.last_post_added_at: datetime = datetime.fromtimestamp(self.last_post_added_ut, timezone.utc)
        ("")

        self.display_layout: str = d['display_layout'] or ''
        ("""
            Either: empty string, `TIMELINE`, or `GALLERY`.

            Value is an empty string on new collections. Empty string is treated the same as `TIMELINE`.
            """)
        self.layout: str = 'TIMELINE'
        ("""
            Either `TIMELINE` or `GALLERY`.

            This is the same as the :attr:`display_layout` attribute but the empty string is changed to `TIMELINE`.
            """)
        if self.display_layout:
            self.layout = self.display_layout

        submission_full_id36s: Sequence[str] = d['link_ids']
        self.submission_id36s: Sequence[str] = [s[3:] for s in submission_full_id36s]
        ("""
            List of submission ID36s contained in the collection.
            """)
        self.submission_ids: Sequence[int] = [int(s, 36) for s in self.submission_id36s]
        ("""
            List of submission IDs contained in the collection.
            """)

class SubmissionCollection(SubmissionCollectionInfo, Sequence[Submission]):
    @property
    def submissions(self) -> Sequence[Submission]:
        """List of submission objects contained in the collection."""
        return self.__submissions

    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.__submissions: Sequence[Submission] = ()
        # https://github.com/python/mypy/issues/4177
        if self.__class__.submissions == __class__.submissions:  # type: ignore[name-defined]
            self.__submissions = self._load_submissions(d, load_submission)

    def __len__(self) -> int:
        return len(self.__submissions)

    def __contains__(self, item: object) -> bool:
        return item in self.__submissions

    def __iter__(self) -> Iterator[Submission]:
        return iter(self.__submissions)

    @overload
    def __getitem__(self, index: int) -> Submission: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[Submission]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[Submission, Sequence[Submission]]:
        return self.__submissions[index]

    _TSubmission = TypeVar('_TSubmission', bound=Submission)

    @final
    def _load_submissions(self,
        d: Mapping[str, Any],
        load: Callable[[Mapping[str, Any]], _TSubmission],
    ) -> Sequence[_TSubmission]:
        children_data = d['sorted_links']['data']['children']
        return [load(i['data']) for i in children_data]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Mapping
if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from ...events import Frame

import asyncio

import aiohttp  # type: ignore[import]

from ..reg_ASYNC import register
from ...websocket_ASYNC import PartiallyImplementedWebSocket, DEFAULT_WAITTIME
from ... import exceptions
from ... import events
from ...const import Side, ConnectionState

class WebSocketClient(PartiallyImplementedWebSocket):
    SIDE: int = Side.CLIENT

    def __init__(self, ws: aiohttp.ClientWebSocketResponse, session: aiohttp.ClientSession) -> None:
        super().__init__()
        self.ws: aiohttp.ClientWebSocketResponse = ws
        ("")
        self.session: aiohttp.ClientSession = session
        ("")

    async def send_frame(self, m: Frame) -> None:
        raise RuntimeError('sending raw frames is not supported by aiohttp')

    async def send_text(self, data: str) -> None:
        if self.state == ConnectionState.CLOSED:
            raise exceptions.ConnectionClosedException
        if self.state != ConnectionState.OPEN:
            raise exceptions.InvalidStateException(f'cannot send frame in {self.state.name!r} state')

        try:
            await self.ws.send_str(data)
        except Exception as cause:
            raise exceptions.TransportError from cause

    async def send_bytes(self, data: bytes) -> None:
        if self.state == ConnectionState.CLOSED:
            raise exceptions.ConnectionClosedException
        if self.state != ConnectionState.OPEN:
            raise exceptions.InvalidStateException(f'cannot send frame in {self.state.name!r} state')

        try:
            await self.ws.send_bytes(data)
        except Exception as cause:
            raise exceptions.TransportError from cause

    async def _pulse_impl(self, *, timeout: float = -2) -> AsyncIterator[object]:
        t: Optional[float] = timeout
        if timeout == -2:
            t = self.default_waittime
        elif timeout == -1:
            t = None
        elif timeout < 0:
            raise ValueError(f'invalid timeout value: {t}')

        try:
            wsm = await self.ws.receive(timeout=t)
        except asyncio.TimeoutError as cause:
            raise exceptions.TimeoutException from cause
        except Exception as cause:
            raise exceptions.TransportError from cause

        if wsm.type == aiohttp.WSMsgType.TEXT:
            yield events.TextMessage(wsm.data)
        elif wsm.type == aiohttp.WSMsgType.BINARY:
            yield events.BytesMessage(wsm.data)
        elif wsm.type == aiohttp.WSMsgType.CLOSING:
            self.state = ConnectionState.CLOSING
        elif wsm.type == aiohttp.WSMsgType.CLOSED:
            close_code = self.ws.close_code
            if not close_code:
                raise Exception('assertion')
            self.close_code: int = close_code
            self.state = ConnectionState.CLOSED
            yield events.ConnectionClosed()
        elif wsm.type == aiohttp.WSMsgType.ERROR:
            raise exceptions.TransportError from wsm.data

    async def close(self, code: Optional[int] = 1000, reason: str = '', *, waitfor: float = -2) -> None:
        if code is None:
            raise RuntimeError('must specify a close code with aiohttp library websockets')

        t: Optional[float] = waitfor
        if waitfor == -2:
            t = self.default_waittime
        elif waitfor == -1:
            t = None
        elif waitfor < 0:
            raise ValueError(f'invalid waitfor value: {waitfor}')

        if self.state == ConnectionState.CLOSED:
            return
        if self.state != ConnectionState.OPEN:
            raise exceptions.InvalidStateException(f'operation not permitted in {self.state.name!r} state')

        self.state: ConnectionState = ConnectionState.CLOSING

        coro = self.ws.close(code=code, message=reason.encode())
        try:
            await asyncio.wait_for(coro, t)
        except asyncio.TimeoutError:
            pass
        except Exception as e:
            raise exceptions.TransportError from e

        if self.ws.close_code:
            self.close_code = self.ws.close_code

        if self.close_code < 0:
            self.close_code = 1006
        self.state = ConnectionState.CLOSED

        await self.session.close()


async def connect(
    url: str,
    *,
    subprotocols: Sequence[str] = (),
    headers: Optional[Mapping[str, str]] = None,
    timeout: float = -2,
) -> WebSocketClient:
    t: Optional[float] = timeout
    if timeout == -2:
        t = DEFAULT_WAITTIME
    elif timeout == -1:
        t = None
    elif timeout < 0:
        raise ValueError(f'invalid timeout value: {timeout}')

    session = aiohttp.ClientSession()
    coro = session.ws_connect(
        url,
        headers=headers,
        protocols=subprotocols,
    )
    try:
        ws = await asyncio.wait_for(coro, t)
    except asyncio.TimeoutError:
        await session.close()
        raise exceptions.TimeoutException

    wsc = WebSocketClient(ws, session)
    wsc.subprotocol = ws.protocol or ''
    return wsc


name: str = aiohttp.__name__
version: str = aiohttp.__version__
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    connect=connect,
)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...http.handler_SYNC import Handler
    from ...http.send_params import SendParams
    from ...http.exchange import Exchange

import time

from ...http.delegating_handler_SYNC import DelegatingHandler
from ...util.token_bucket import TokenBucket

class RateLimited(DelegatingHandler):
    def __init__(self, handler: Handler) -> None:
        super().__init__(handler)
        self._tb = TokenBucket(3, 1/2)

    def _send(self, p: SendParams) -> Exchange:
        time.sleep(self._tb.get_cooldown(1))
        self._tb.consume(1)
        return super()._send(p)




from .pull_async1 import (  # noqa: F401
    OverviewListingAsyncPaginator as OverviewListingAsyncPaginator,
    CommentsListingAsyncPaginator as CommentsListingAsyncPaginator,
    SubmittedListingAsyncPaginator as SubmittedListingAsyncPaginator,
    GildedListingAsyncPaginator as GildedListingAsyncPaginator,
    UpvotedListingAsyncPaginator as UpvotedListingAsyncPaginator,
    DownvotedListingAsyncPaginator as DownvotedListingAsyncPaginator,
    HiddenListingAsyncPaginator as HiddenListingAsyncPaginator,
    SavedListingAsyncPaginator as SavedListingAsyncPaginator,
    SavedSubmissionsListingAsyncPaginator as SavedSubmissionsListingAsyncPaginator,
    SavedCommentsListingAsyncPaginator as SavedCommentsListingAsyncPaginator,
)
from .pull_user_subreddits_async1 import UserSubredditListingAsyncPaginator as UserSubredditListingAsyncPaginator  # noqa: F401
from .explore_async1 import UserSearchAsyncPaginator as UserSearchAsyncPaginator  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.moderation_action_log_entry import ModerationActionLogEntry
    from ..stream_ASYNC import IStandardStreamEventSubject
    from ...pagination.async_paginator import CursorAsyncPaginator

from ..stream_ASYNC import Stream


def action_log_extractor(x: ModerationActionLogEntry) -> object:
    return x.uuid

def get_action_log_stream_paginator(client: Client, sr: str) -> CursorAsyncPaginator[ModerationActionLogEntry]:
    return client.p.moderation.pull_actions(sr).get_paginator()

def make_action_log_stream(paginator: CursorAsyncPaginator[ModerationActionLogEntry], past: Optional[Iterable[ModerationActionLogEntry]] = None, seen: Optional[Iterable[object]] = None) -> IStandardStreamEventSubject[ModerationActionLogEntry]:
    return Stream(paginator, action_log_extractor, past=past, seen=seen)

def create_action_log_stream(client: Client, sr: str) -> IStandardStreamEventSubject[ModerationActionLogEntry]:
    return make_action_log_stream(get_action_log_stream_paginator(client, sr))




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.comment_tree_ASYNC import MoreCommentsTreeNode, SubmissionTreeNode

from ...util.base_conversion import to_base36
from .low_ASYNC import Low

class CommentTreeProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.low: Low = Low(self, client)
        ("""
            Low level calls for efficiency.
            """)

    async def get(self,
        submission_id: int,
        comment_id: Optional[int] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> Optional[SubmissionTreeNode]:
        """Get the comment tree for a submission.

        Behaves similarly to :meth:`.fetch`.

        Returns `None` instead of raises on `StatusCodeException(404)` and `RejectedResultException`.
        """
        submission_id36 = to_base36(submission_id)
        comment_id36 = None if comment_id is None else to_base36(comment_id)
        return await self.low.get(
            submission_id36,
            comment_id36,
            sort=sort,
            limit=limit,
            depth=depth,
            context=context,
        )

    async def fetch(self,
        submission_id: int,
        comment_id: Optional[int] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> SubmissionTreeNode:
        """Get the comment tree for a submission.

        .. .PARAMETERS

        :param `int` submission_id:
            Submission ID.
        :param `Optional[int]` comment_id:
            Optional comment ID to start the tree at that comment.
        :param `str` sort:
            Either: `confidence` ('best'), `top`, `new`, `controversial`, `old`, `random`, `qa`, `live`.

            If not given or not a valid sort value (including empty string), the default is the
            'sort comments by' preference of the logged in user. Otherwise, if there is no user
            context the default is `confidence`.
        :param `Optional[int]` limit:
            Limit the number of comments to retrieve.

            The effective default seems to be 200, and the max value appears to be 500.
        :param `Optional[int]` depth:
            The number of levels deep to retrieve comments for.

            A value of 0 is ignored.
            A value of 1 means to only retrieve top-level comments.
            A value of 2 means to retrieve comments one level deep.
            And so on.

            The maximum is 10, which is also the default.
            Any value higher than 10 is treated the same as 10.
        :param `Optional[int]` context:
            If `comment_id` is specified, the number of parent comments to include.

            Specify an integer from 0 to 8. Any number higher than 8 is treated the same as 8.

        .. .RETURNS

        :rtype: :class:`~.models.comment_tree_ASYNC.SubmissionTreeNode`

        .. .RAISES

        :raises redditwarp.exceptions.RejectedResultException:
            When specifying a comment using the `comment_id` parameter and the returned
            comment list is empty, so the specified comment was not returned.

            This happens when the comment existed at one point but is no longer available
            anymore and no trace of the comment exists in the comment tree whatsoever.
            This can occur when a deleted comment has no replies but your program still
            has a reference to the comment and tried to retrieve it.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
               - The specified submission ID does not exist.
               - The specified comment ID does not exist or the comment belongs
                 to a submission other than the one specified.
        """
        submission_id36 = to_base36(submission_id)
        comment_id36 = None if comment_id is None else to_base36(comment_id)
        return await self.low.fetch(
            submission_id36,
            comment_id36,
            sort=sort,
            limit=limit,
            depth=depth,
            context=context,
        )

    async def fetch_lenient(self,
        submission_id: int,
        comment_id: Optional[int] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> SubmissionTreeNode:
        """Get the comment tree for a submission.

        Behaves similarly to :meth:`.fetch`.

        This method does the same thing as :meth:`.fetch` but doesn't reject
        with a `RejectedResultException` when the specified comment ID could
        not be retrieved.
        """
        submission_id36 = to_base36(submission_id)
        comment_id36 = None if comment_id is None else to_base36(comment_id)
        return await self.low.fetch(
            submission_id36,
            comment_id36,
            sort=sort,
            limit=limit,
            depth=depth,
            context=context,
        )

    async def more_children(self,
        submission_id: int,
        child_ids: Iterable[int],
        *,
        sort: str = '',
        depth: Optional[int] = None,
        exact: bool = False,
    ) -> MoreCommentsTreeNode:
        """Retrieve comments omitted from a comment tree.

        When a comment tree is rendered, the most relevant comments are selected
        for display and the remaining comments are stubbed out with more-comment
        links: either 'load more comments' or 'continue this thread'. This procedure
        is used to retrieve the comments represented by the 'load more comments' stubs.

        You may only make one request at a time to this API endpoint.
        Higher concurrency will result in an error being returned.

        .. .PARAMETERS

        :param `int` submission_id:
        :param `Iterable[int]` child_ids:
            A list of comment IDs.
        :param `str` sort:
            Same as on :meth:`.fetch`.
        :param `Optional[int]` depth:
            Same as on :meth:`.fetch`.
        :param `bool` exact:
            If true, only return the children requested, and not their sub-comments.

            This is kind of the same as specifying `depth=1` but more-comment objects won't be present.

            If this is specified with the `depth` parameter, this parameter will take precedence.

        .. .RETURNS

        :rtype: :class:`~.models.comment_tree_ASYNC.MoreCommentsTreeNode`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The specified submission does not exist.
        """
        submission_id36 = to_base36(submission_id)
        child_id36s = (to_base36(x) for x in child_ids)
        return await self.low.more_children(
            submission_id36,
            child_id36s,
            sort=sort,
            depth=depth,
            exact=exact,
        )




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, Iterable
if TYPE_CHECKING:
    from ...client_SYNC import Client

from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_calling_iterator import CallChunkCallingIterator
from ...iterators.call_chunk_chaining_iterator import CallChunkChainingIterator
from ...iterators.call_chunk import CallChunk
from ...models.modmail_SYNC import (
    ConversationInfo,
    Message,
    ConversationAggregate,
    UserDossierConversationAggregate,
    OptionalUserDossierConversationAggregate,
)
from ...model_loaders.modmail_SYNC import (
    load_conversation_info,
    load_message,
    load_conversation_aggregate,
    load_user_dossier_conversation_aggregate,
    load_optional_user_dossier_conversation_aggregate,
)

class ConversationProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    def fetch(self, idn: int, *, mark_read: bool = False) -> OptionalUserDossierConversationAggregate:
        """Get a conversation.

        .. .PARAMETERS

        :param `int` idn:
            Conversation ID.
        :param `bool` mark_read:
            Mark retrieved conversations as read.

            Default: false.

        .. .RETURNS

        :rtype: :class:`~.models.modmail_SYNC.OptionalUserDossierConversationAggregate`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `SUBREDDIT_NO_ACCESS`:
                The subreddit associated with the conversation is not moderated by you.
            + `CONVERSATION_NOT_FOUND`:
                The specified conversation does not exist.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('GET', f'/api/mod/conversations/{convo_id36}')
        return load_optional_user_dossier_conversation_aggregate(
            root['conversation'],
            root['messages'],
            root['modActions'],
            root['user'] or None,
            client=self._client,
        )

    def create(self, sr: str, to: str, subject: str, body: str, *, hidden: bool = False) -> tuple[ConversationInfo, Message]:
        """Get a conversation.

        .. .PARAMETERS

        :param `str` sr:
            The name of the subreddit in which to create the conversation for.
        :param `str` to:
            The modmail recipient name.

            To create a moderator conversation, specify an empty string value.

            If the specified user is a moderator of the subreddit, this parameter
            is ignored and an internal moderator conversation is created instead.
        :param `str` subject:
            A subject line for the conversation.
        :param `str` body:
            Markdown text.
        :param `bool` hidden:
            Expose your user name to the recipient.

            Default: false.

        .. .RETURNS

        :rtype: `tuple`\\[:class:`~.models.modmail_SYNC.ConversationInfo`, :class:`~.models.modmail_SYNC.Message`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `BAD_SR_NAME`:
                The specified subreddit was empty.
            + `SUBREDDIT_NOEXIST`:
                The specified subreddit does not exist.
            + `NO_TEXT`:
               - The `subject` was empty.
               - The `body` was empty.

            + `TOO_LONG`:
               - The value specified for `subject` must be 100 characters or fewer
                 (despite error message saying under 100).
               - The value specified for `body` must be 10000 characters or fewer
                 (despite error message saying under 10000).
        """
        data = {'srName': sr, 'to': to, 'subject': subject, 'body': body, 'isAuthorHidden': '01'[hidden]}
        root = self._client.request('POST', '/api/mod/conversations', data=data)
        conversation_data = root['conversation']
        conversation = load_conversation_info(conversation_data, self._client)
        message_id36 = conversation_data['objIds'][0]['id']
        message = load_message(root['messages'][message_id36], self._client)
        return (conversation, message)

    def reply(self, idn: int, body: str, *, hidden: bool = False, internal: bool = False) -> ConversationAggregate:
        """Create a new message on an existing conversation.

        .. .PARAMETERS

        :param `int` idn:
            Conversation ID.
        :param `str` body:
            Markdown text.
        :param `bool` hidden:
            Expose your user name to the recipient.

            Default: false.
        :param `bool` internal:
            Create a private moderator note.

            Default: false.

        .. .RETURNS

        :rtype: :class:`~.models.modmail_SYNC.ConversationAggregate`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `CONVERSATION_NOT_FOUND`:
                The specified conversation does not exist.
            + `SUBREDDIT_NO_ACCESS`:
                The subreddit associated with the conversation is not moderated by you.
            + `NO_TEXT`:
                The `body` was empty.
            + `TOO_LONG`:
                The value specified for `body` must be 10000 characters or fewer
                (despite error message saying under 10000).
        """
        data = {'body': body, 'isAuthorHidden': '01'[hidden], 'isInternal': '01'[internal]}
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}', data=data)
        conversation_data = root['conversation']
        messages_mapping_data = root['messages']
        mod_actions_mapping_data = conversation_data['modActions']
        return load_conversation_aggregate(
            conversation_data,
            messages_mapping_data,
            mod_actions_mapping_data,
            client=self._client,
        )

    def mark_read(self, idn: int) -> None:
        """Mark a conversation as read.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        data = {'conversationIds': to_base36(idn)}
        self._client.request('POST', '/api/mod/conversations/read', data=data)

    def bulk_mark_read(self, ids: Iterable[int]) -> CallChunkCallingIterator[None]:
        """Mark conversations as read.

        Any specified conversation that does not exist will be ignored.
        If any of the IDs refer to a conversation you do not have permission over,
        an `INVALID_CONVERSATION_ID` API error will occur and none of the conversations
        will be processed.

        .. .PARAMETERS

        :param `Iterable[int]` ids:

        .. .RETURNS

        :rtype: :class:`~.iterators.call_chunk_calling_iterator.CallChunkCallingIterator`\\[`None`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `INVALID_CONVERSATION_ID`:
                You do not have permission to mark as read one of the specified conversations.
                The operation is aborted and none of the items will be processed.
        """
        def mass_mark_read(ids: Sequence[int]) -> None:
            id36s = map(to_base36, ids)
            ids_str = ','.join(id36s)
            self._client.request('POST', '/api/mod/conversations/read', data={'conversationIds': ids_str})

        return CallChunkCallingIterator(CallChunk(mass_mark_read, chunk) for chunk in chunked(ids, 100))

    def mark_unread(self, idn: int) -> None:
        """Mark a conversation as unread.

        Behaves similarly to :meth:`.mark_read`.
        """
        data = {'conversationIds': to_base36(idn)}
        self._client.request('POST', '/api/mod/conversations/unread', data=data)

    def bulk_mark_unread(self, ids: Iterable[int]) -> CallChunkCallingIterator[None]:
        """Mark conversations as unread.

        Behaves similarly to :meth:`.bulk_mark_read`.
        """
        def mass_mark_unread(ids: Sequence[int]) -> None:
            id36s = map(to_base36, ids)
            ids_str = ','.join(id36s)
            self._client.request('POST', '/api/mod/conversations/unread', data={'conversationIds': ids_str})

        return CallChunkCallingIterator(CallChunk(mass_mark_unread, chunk) for chunk in chunked(ids, 100))

    def bulk_mark_all_read(self, mailbox: str, subrs: Iterable[str]) -> CallChunkChainingIterator[int]:
        """Mark all conversations across select mailboxes and subreddits as read.

        Specified subreddit names that do not exist will be ignored, but if none of
        the subreddits exist then a 500 HTTP error will occur. If any of the subreddits
        are not moderated by you then a `BAD_SR_NAME` API error will occur, and none of
        the conversations will be processed.

        .. .PARAMETERS

        :param `str` mailbox:
            Either: `all`, `appeals`, `notifications`, `inbox`, `filtered`, `inprogress`,
            `mod`, `archived`, `default`, `highlighted`, `join_requests`, `new`.

            Defaults to `all` if empty string.
        :param `Iterable[str]` subrs:
            Subreddit names.

        .. .RETURNS

        :rtype: :class:`~.iterators.call_chunk_chaining_iterator.CallChunkChainingIterator`\\[`int`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `INVALID_OPTION`
                The specified `mailbox` was invalid.
            + `BAD_SR_NAME`:
                One of the specified subreddits is not a subreddit you have access to.
        """
        def mass_mark_all_read(subrs: Sequence[str]) -> Sequence[int]:
            subrs_str = ','.join(subrs)
            data = {'state': mailbox, 'entity': subrs_str}
            root = self._client.request('POST', '/api/mod/conversations/bulk/read', data=data)
            return [int(id36, 36) for id36 in root['conversation_ids']]

        return CallChunkChainingIterator(CallChunk(mass_mark_all_read, chunk) for chunk in chunked(subrs, 100))

    def highlight(self, idn: int) -> ConversationAggregate:
        """Mark a conversation as highlighted.

        .. .PARAMETERS

        :params `int` idn:

        .. .RETURNS

        :rtype: :class:`~.models.modmail_SYNC.ConversationAggregate`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `CONVERSATION_NOT_FOUND`:
                The specified conversation does not exist.
            + `SUBREDDIT_NO_ACCESS`:
                The subreddit associated with the conversation is not moderated by you.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/highlight')
        return load_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            client=self._client,
        )

    def unhighlight(self, idn: int) -> ConversationAggregate:
        """Unmark a conversation as highlighted.

        Behaves similarly to :meth:`.highlight`.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('DELETE', f'/api/mod/conversations/{convo_id36}/highlight')
        return load_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            client=self._client,
        )

    def archive(self, idn: int) -> ConversationAggregate:
        """Archive a conversation.

        .. hint::
            Archiving a conversation moves it to the `archived` mailbox folder.

        .. .PARAMETERS

        :params `int` idn:

        .. .RETURNS

        :rtype: :class:`~.models.modmail_SYNC.ConversationAggregate`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `CONVERSATION_NOT_FOUND`:
                The specified conversation does not exist.
            + `INVALID_MOD_PERMISSIONS`:
                The subreddit associated with the conversation is not moderated by you.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/archive')
        return load_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            client=self._client,
        )

    def unarchive(self, idn: int) -> ConversationAggregate:
        """Unmark a conversation as highlighted.

        Behaves similarly to :meth:`.highlight`.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/unarchive')
        return load_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            client=self._client,
        )

    def approve_user(self, idn: int) -> UserDossierConversationAggregate:
        """Approve the user associated with a conversation.

        .. .PARAMETERS

        :params `int` idn:

        .. .RETURNS

        :rtype: :class:`~.models.modmail_SYNC.UserDossierConversationAggregate`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `CONVERSATION_NOT_FOUND`:
                The specified conversation does not exist.
            + `INVALID_MOD_PERMISSIONS`:
                The subreddit associated with the conversation is not moderated by you.
            + `CANT_RESTRICT_MODERATOR`:
                There is no user associated with the conversation.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/approve')
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    def unapprove_user(self, idn: int) -> UserDossierConversationAggregate:
        """Unapprove the user associated with a conversation.

        Behaves similarly to :meth:`.approve_user`.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/disapprove')
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    def mute_user_3d(self, idn: int) -> UserDossierConversationAggregate:
        """Mute the user associated with a conversation for 3 days.

        .. .PARAMETERS

        :params `int` idn:

        .. .RETURNS

        :rtype: :class:`~.models.modmail_SYNC.UserDossierConversationAggregate`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `CONVERSATION_NOT_FOUND`:
                The specified conversation does not exist.
            + `INVALID_MOD_PERMISSIONS`:
                The subreddit associated with the conversation is not moderated by you.
            + `CANT_RESTRICT_MODERATOR`:
                There is no user associated with the conversation.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/mute', data={'num_hours': '72'})
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    def mute_user_7d(self, idn: int) -> UserDossierConversationAggregate:
        """Mute the user associated with a conversation for 7 days.

        Behaves similarly to :meth:`.mute_user_3d`.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/mute', data={'num_hours': '168'})
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    def mute_user_28d(self, idn: int) -> UserDossierConversationAggregate:
        """Mute the user associated with a conversation for 28 days.

        Behaves similarly to :meth:`.mute_user_3d`.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/mute', data={'num_hours': '672'})
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    def unmute_user(self, idn: int) -> UserDossierConversationAggregate:
        """Unmute the user associated with a conversation.

        Behaves similarly to :meth:`.mute_user_3d`.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/unmute')
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    def shorten_user_ban(self, idn: int, duration: int) -> UserDossierConversationAggregate:
        """Switch a permanent ban to a temporary one of the user associated with a conversation.

        If the user is not permanently banned, an API error will be raised.

        .. .PARAMETERS

        :param `int` idn:
        :param `int` duration:
            The number of days the temporary ban should last.

            Specify an integer from 1 to 999.

            The UI has the options: 1, 3, 7, or 28 days.

        .. .RETURNS

        :rtype: :class:`~.models.modmail_SYNC.UserDossierConversationAggregate`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `CONVERSATION_NOT_FOUND`:
                The specified conversation does not exist.
            + `INVALID_MOD_PERMISSIONS`:
                The subreddit associated with the conversation is not moderated by you.
            + `CANT_RESTRICT_MODERATOR`:
                There is no user associated with the conversation.
            + `Participant must be banned.`:
                The user associated with the conversation is not banned from the subreddit.
            + `Participant must be banned permanently.`:
                The user associated with the conversation is not permanently banned from the subreddit.
            + `BAD_NUMBER`:
                The number specified by the `duration` parameter was not in range.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/temp_ban',
                data={'duration': str(duration)})
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    def unban_user(self, idn: int) -> UserDossierConversationAggregate:
        """Unban the user associated with a conversation from the subreddit.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: :class:`~.models.modmail_SYNC.UserDossierConversationAggregate`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `CONVERSATION_NOT_FOUND`:
                The specified conversation does not exist.
            + `INVALID_MOD_PERMISSIONS`:
                The subreddit associated with the conversation is not moderated by you.
            + `CANT_RESTRICT_MODERATOR`:
                There is no user associated with the conversation.
        """
        convo_id36 = to_base36(idn)
        root = self._client.request('POST', f'/api/mod/conversations/{convo_id36}/unban')
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, MutableMapping
if TYPE_CHECKING:
    from ...http.handler_ASYNC import Handler

from ...core.http_client_ASYNC import HTTPClient
from ...core.rate_limited_ASYNC import RateLimited
from ...core.reddit_please_send_json_ASYNC import RedditPleaseSendJSON
from ...http.misc.apply_params_and_headers_ASYNC import ApplyDefaultHeaders
from ...http.transport.reg_ASYNC import new_connector
from ...http.util.case_insensitive_dict import CaseInsensitiveDict
from ..auth.token_obtainment_client_ASYNC import new_token_obtainment_client
from .authorizer_ASYNC import Authorizer, Authorized
from ...core.user_agent_ASYNC import get_user_agent
from ...http.http_client_ASYNC import HTTPClient as BaseHTTPClient
from ...core.direct_by_origin_ASYNC import DirectByOrigin
from .const import TRUSTED_ORIGINS


class RedditHTTPClient(HTTPClient):
    @property
    def authorizer(self) -> Authorizer:
        return self.fetch_authorizer()

    def __init__(self, handler: Handler, *,
        headers: Optional[MutableMapping[str, str]] = None,
        authorizer: Optional[Authorizer] = None,
    ) -> None:
        super().__init__(handler, headers=headers)
        self._authorizer: Optional[Authorizer] = authorizer

    def fetch_authorizer(self) -> Authorizer:
        if self._authorizer is None:
            raise RuntimeError('value not set')
        return self._authorizer

    def fast_set_authorizer(self, value: Optional[Authorizer]) -> None:
        self._authorizer = value


def build_reddit_http_client() -> RedditHTTPClient:
    connector = new_connector()
    ua = get_user_agent(module_member=connector)
    headers = CaseInsensitiveDict({'User-Agent': ua})
    http = BaseHTTPClient(ApplyDefaultHeaders(connector, headers))
    token_client = new_token_obtainment_client(http)
    authorizer = Authorizer(token_client)
    handler: Handler
    handler = RedditPleaseSendJSON(RateLimited(Authorized(connector, authorizer)))
    directions = {x: handler for x in TRUSTED_ORIGINS}
    handler = DirectByOrigin(connector, directions)
    http = RedditHTTPClient(handler, headers=headers, authorizer=authorizer)
    return http




from __future__ import annotations
from typing import Optional, Any

import requests  # type: ignore[import]
import requests.adapters  # type: ignore[import]

from ... import exceptions
from ... import payload
from ...send_params import SendParams
from ...exchange import Exchange
from ...request import Request
from ...response import UResponse
from ..reg_SYNC import register
from ..connector_SYNC import Connector


def _get_effective_timeout(v: float) -> float:
    if 0 > v != -1:
        raise ValueError(f"invalid `timeout` value: {v}")
    return v

def _get_effective_follow_redirects(v: Optional[bool]) -> bool:
    if v is None:
        raise ValueError(f"invalid `follow_redirects` value: {v}")
    return v


class RequestsConnector(Connector):
    def __init__(self,
        session: requests.Session,
    ) -> None:
        super().__init__()
        self.session: requests.Session = session
        ("")

    def _send(self, p: SendParams) -> Exchange:
        r = p.requisition

        timeout = _get_effective_timeout(p.timeout)
        follow_redirects = _get_effective_follow_redirects(p.follow_redirects)

        headers: dict[str, str] = dict(r.headers)
        data: object = None
        json: object = None
        files: Any = None

        pld = r.payload
        if pld is None:
            pass

        elif isinstance(pld, payload.Bytes):
            headers['Content-Type'] = pld.get_media_type()
            data = pld.data

        elif isinstance(pld, payload.Text):
            headers['Content-Type'] = pld.get_media_type()
            data = pld.text.encode()

        elif isinstance(pld, payload.JSON):
            json = pld.json

        elif isinstance(pld, payload.URLEncodedFormData):
            data = pld.data

        elif isinstance(pld, payload.MultipartFormData):
            files = {}
            for pt in pld.parts:
                if isinstance(pt, payload.MultipartFormData.TextField):
                    files[pt.name] = (None, pt.text)
                elif isinstance(pt, payload.MultipartFormData.FileField):
                    files[pt.name] = (pt.filename, pt.file, pt.content_type)
                else:
                    raise ValueError('unexpected multipart field type: ' + repr(pt))

        else:
            raise Exception(f"unsupported payload type: {pld.__class__.__name__!r}")

        try:
            resp = self.session.request(
                method=r.verb,
                url=r.url,
                params=r.params,
                headers=headers,
                timeout=timeout,
                allow_redirects=follow_redirects,
                data=data,
                json=json,
                files=files,
            )
        except requests.exceptions.ReadTimeout as cause:
            raise exceptions.TimeoutException from cause
        except Exception as cause:
            raise exceptions.TransportError from cause

        requ = resp.request

        x_requ_data = requ.body
        if isinstance(x_requ_data, str):
            x_requ_data = x_requ_data.encode()
        elif x_requ_data is None:
            x_requ_data = b''

        x_requ = Request(
            verb=requ.method or '',
            url=requ.url or '',
            headers=requ.headers,
            data=x_requ_data,
        )
        x_resp = UResponse(
            status=resp.status_code,
            headers=resp.headers,
            data=resp.content,
            underlying_object=resp,
        )
        history = [
            UResponse(
                status=resp1.status_code,
                headers=resp1.headers,
                data=resp1.content,
                underlying_object=resp1,
            )
            for resp1 in resp.history
        ]
        return Exchange(
            requisition=r,
            request=x_requ,
            response=x_resp,
            history=history,
        )

    def _close(self) -> None:
        self.session.close()


def new_connector() -> RequestsConnector:
    return RequestsConnector(requests.Session())


name: str = requests.__name__
version: str = requests.__version__
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    new_connector=new_connector,
)




from __future__ import annotations
from typing import Any, Mapping

from ..models.trophy import Trophy

def load_trophy(d: Mapping[str, Any]) -> Trophy:
    return Trophy(
        d=d,
        name=d['name'],
        description=d['description'] or '',
        icon_40=d['icon_40'],
        icon_70=d['icon_70'],
    )




from __future__ import annotations
from typing import Any, Mapping

from ..models.user_summary import UserSummary

def load_user_summary(d: Mapping[str, Any], id36: str) -> UserSummary:
    return UserSummary(d, id36)




from __future__ import annotations
from typing import Sequence

from .listing.listing_paginator import ListingPaginator
from ...models.live_thread_SYNC import LiveUpdate
from ...model_loaders.live_thread_SYNC import load_live_update

class LiveUpdateListingPaginator(ListingPaginator[LiveUpdate]):
    def fetch(self) -> Sequence[LiveUpdate]:
        data = self._fetch_data()
        return [load_live_update(d['data'], self.client) for d in data['children']]




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Optional
if TYPE_CHECKING:
    from ..client_ASYNC import Client
    from .comment_ASYNC import Comment

from .submission import (
    Submission as BaseSubmission,
    LinkPost as BaseLinkPost,
    TextPost as BaseTextPost,
    GalleryPost as BaseGalleryPost,
    PollPost as BasePollPost,
    CrosspostSubmission as BaseCrosspostSubmission,
)

class Submission(BaseSubmission):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")

    async def reply(self, text: str) -> Comment:
        return await self.client.p.submission.reply(self.id, text)

    async def delete(self) -> None:
        await self.client.p.submission.delete(self.id)

    async def lock(self) -> None:
        await self.client.p.submission.lock(self.id)

    async def unlock(self) -> None:
        await self.client.p.submission.unlock(self.id)

    async def distinguish(self) -> Submission:
        return await self.client.p.submission.distinguish(self.id)

    async def undistinguish(self) -> Submission:
        return await self.client.p.submission.undistinguish(self.id)

    async def sticky(self, slot: Optional[int] = None) -> None:
        await self.client.p.submission.sticky(self.id, slot)

    async def unsticky(self) -> None:
        await self.client.p.submission.unsticky(self.id)

    async def approve(self) -> None:
        await self.client.p.submission.approve(self.id)

    async def remove(self) -> None:
        await self.client.p.submission.remove(self.id)

    async def apply_removal_reason(self,
            reason_id: Optional[str],
            note: Optional[str] = None) -> None:
        await self.client.p.submission.apply_removal_reason(self.id, reason_id, note)

    async def send_removal_comment(self,
            title: str,
            message: str) -> Comment:
        return await self.client.p.submission.send_removal_comment(self.id, title, message)

    async def send_removal_message(self,
            title: str,
            message: str,
            *,
            exposed: bool = False) -> None:
        await self.client.p.submission.send_removal_message(self.id, title, message, exposed=exposed)


class LinkPost(Submission, BaseLinkPost):
    pass

class TextPost(Submission, BaseTextPost):
    async def edit_body(self, text: str) -> TextPost:
        return await self.client.p.submission.edit_text_post_body(self.id, text)

class GalleryPost(Submission, BaseGalleryPost):
    pass

class PollPost(Submission, BasePollPost):
    pass

class CrosspostSubmission(Submission, BaseCrosspostSubmission):
    @property
    def original(self) -> Optional[Submission]:
        return self.__original

    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d, client)
        from ..model_loaders.submission_ASYNC import load_submission  # Avoid cyclic import
        load = lambda d: load_submission(d, client)
        self.__original: Optional[Submission] = self._load_original(d, load)




from __future__ import annotations
from typing import Optional, Any, Mapping, Iterator
from dataclasses import dataclass, field

@dataclass(frozen=True)
class Token(Mapping[str, object]):
    """OAuth2 token information."""

    access_token: str
    #_: KW_ONLY
    token_type: str = 'Bearer'
    expires_in: Optional[int] = None
    refresh_token: Optional[str] = None
    scope: Optional[str] = None
    d: Mapping[str, object] = field(repr=False, default_factory=dict)

    @staticmethod
    def from_dict(d: Mapping[str, Any]) -> Token:
        """Create an instance from OAuth2 token data."""
        return Token(
            access_token=d['access_token'],
            token_type=d['token_type'],
            expires_in=d.get('expires_in'),
            refresh_token=d.get('refresh_token'),
            scope=d.get('scope'),
            d=d,
        )

    def __contains__(self, item: object) -> bool:
        return item in self.d
    def __iter__(self) -> Iterator[str]:
        return iter(self.d)
    def __len__(self) -> int:
        return len(self.d)
    def __getitem__(self, key: str) -> object:
        return self.d[key]




from __future__ import annotations
from typing import Any, Mapping, Optional

from datetime import datetime, timezone

from ..models.submission_draft import (
    SubmissionDraft,
    MarkdownTextPostDraft,
    RichTextTextPostDraft,
    LinkPostDraft,
)


normal_to_public_draft_field_names_map: Mapping[str, str] = {
    'id': 'id',
    'kind': 'kind',
    'created': 'created',
    'modified': 'modified',
    'is_public_link': 'isPublicLink',
    'subreddit': 'subredditId',
    'title': 'title',
    'body': 'body',
    'send_replies': 'sendReplies',
    'spoiler': 'isSpoiler',
    'nsfw': 'isNSFW',
    'original_content': 'isOriginalContent',
    'flair': 'flair',
}
normal_draft_field_names_map: Mapping[str, str] = {k: k for k, _ in normal_to_public_draft_field_names_map.items()}
public_draft_field_names_map: Mapping[str, str] = normal_to_public_draft_field_names_map


def _xload_submission_draft_impl(d: Mapping[str, Any], h: Mapping[str, str]) -> SubmissionDraft:
    created_ts: float = d[h['created']] / 1000
    modified_ts: float = d[h['modified']] / 1000

    subreddit_id: Optional[int] = None
    subreddit: Optional[str] = d[h['subreddit']]
    if subreddit is not None:
        _, _, id36 = subreddit.partition('_')
        subreddit_id = int(id36, 36)

    flair: Optional[SubmissionDraft.Flair] = None
    if df := d[h['flair']]:
        flair = SubmissionDraft.Flair(
            uuid=df['templateId'],
            text_mode=df['type'],
            text=df['text'],
            bg_color=df['backgroundColor'],
            fg_color_scheme=df['textColor'],
        )

    return SubmissionDraft(
        d=d,
        uuid=d[h['id']],
        created_at=datetime.fromtimestamp(created_ts, timezone.utc),
        modified_at=datetime.fromtimestamp(modified_ts, timezone.utc),
        public=d[h['is_public_link']],
        subreddit_id=subreddit_id,
        title=d[h['title']],
        reply_notifications=d[h['send_replies']],
        spoiler=d[h['spoiler']],
        nsfw=d[h['nsfw']],
        oc=d[h['original_content']],
        flair=flair,
    )


def _load_markdown_text_post_draft_impl(d: Mapping[str, Any], h: Mapping[str, str]) -> MarkdownTextPostDraft:
    up = _xload_submission_draft_impl(d, h)
    return MarkdownTextPostDraft(
        d=up.d,
        uuid=up.uuid,
        created_at=up.created_at,
        modified_at=up.modified_at,
        public=up.public,
        subreddit_id=up.subreddit_id,
        title=up.title,
        reply_notifications=up.reply_notifications,
        spoiler=up.spoiler,
        nsfw=up.nsfw,
        oc=up.oc,
        flair=up.flair,
        body=d['body'],
    )

def _load_richtext_text_post_draft_impl(d: Mapping[str, Any], h: Mapping[str, str]) -> RichTextTextPostDraft:
    up = _xload_submission_draft_impl(d, h)
    return RichTextTextPostDraft(
        d=up.d,
        uuid=up.uuid,
        created_at=up.created_at,
        modified_at=up.modified_at,
        public=up.public,
        subreddit_id=up.subreddit_id,
        title=up.title,
        reply_notifications=up.reply_notifications,
        spoiler=up.spoiler,
        nsfw=up.nsfw,
        oc=up.oc,
        flair=up.flair,
    )

def _load_link_post_draft_impl(d: Mapping[str, Any], h: Mapping[str, str]) -> LinkPostDraft:
    up = _xload_submission_draft_impl(d, h)
    return LinkPostDraft(
        d=up.d,
        uuid=up.uuid,
        created_at=up.created_at,
        modified_at=up.modified_at,
        public=up.public,
        subreddit_id=up.subreddit_id,
        title=up.title,
        reply_notifications=up.reply_notifications,
        spoiler=up.spoiler,
        nsfw=up.nsfw,
        oc=up.oc,
        flair=up.flair,
        link=d['body'],
    )


def _load_submission_draft_impl(d: Mapping[str, Any], h: Mapping[str, str]) -> SubmissionDraft:
    kind: str = d[h['kind']]
    if kind == 'markdown':
        return _load_markdown_text_post_draft_impl(d, h)
    if kind == 'richtext':
        return _load_richtext_text_post_draft_impl(d, h)
    if kind == 'link':
        return _load_link_post_draft_impl(d, h)
    raise ValueError('unknown draft type')



def load_submission_draft(d: Mapping[str, Any]) -> SubmissionDraft:
    return _load_submission_draft_impl(d, normal_draft_field_names_map)

def load_markdown_text_post_draft(d: Mapping[str, Any]) -> MarkdownTextPostDraft:
    return _load_markdown_text_post_draft_impl(d, normal_draft_field_names_map)

def load_richtext_text_post_draft(d: Mapping[str, Any]) -> RichTextTextPostDraft:
    return _load_richtext_text_post_draft_impl(d, normal_draft_field_names_map)

def load_link_post_draft(d: Mapping[str, Any]) -> LinkPostDraft:
    return _load_link_post_draft_impl(d, normal_draft_field_names_map)


def load_public_submission_draft(d: Mapping[str, Any]) -> SubmissionDraft:
    return _load_submission_draft_impl(d, public_draft_field_names_map)

def load_public_markdown_text_post_draft(d: Mapping[str, Any]) -> MarkdownTextPostDraft:
    return _load_markdown_text_post_draft_impl(d, public_draft_field_names_map)

def load_public_rich_text_text_post_draft(d: Mapping[str, Any]) -> RichTextTextPostDraft:
    return _load_richtext_text_post_draft_impl(d, public_draft_field_names_map)

def load_public_link_post_draft(d: Mapping[str, Any]) -> LinkPostDraft:
    return _load_link_post_draft_impl(d, public_draft_field_names_map)




from __future__ import annotations
from typing import Any, Mapping

from datetime import datetime, timezone

from ..models.moderation_action_log_entry import ModerationActionLogEntry

def load_moderation_action_log_entry(d: Mapping[str, Any]) -> ModerationActionLogEntry:
    return ModerationActionLogEntry(
        d=d,
        uuid=d['id'].partition('_')[-1],
        unixtime=(unixtime := int(d['created_utc'])),
        datetime=datetime.fromtimestamp(unixtime, timezone.utc),
        action=d['action'],
        agent_id=int(d['mod_id36'], 36),
        agent=d['mod'],
    )




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Mapping

from dataclasses import dataclass

from .exceptions import is_successful_status, raise_now, ensure_successful_status


@dataclass(repr=False, eq=False)
class Response:
    status: int
    headers: Mapping[str, str]
    data: bytes

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} [{self.status}]>"

    def is_successful_status(self) -> bool:
        """Return true if `200 <= self.status <= 299`."""
        return is_successful_status(self.status)

    def raise_now(self) -> None:
        """Raises a :class:`~.http.exceptions.StatusCodeException` exception type based on the `.status` number."""
        raise_now(self.status)

    def ensure_successful_status(self) -> None:
        """Raises a :class:`~.http.exceptions.StatusCodeException` exception if `.is_successful_status()` returns false."""
        ensure_successful_status(self.status)


@dataclass(repr=False, eq=False)
class UResponse(Response):
    underlying_object: object = None




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_SYNC import Client

from ..models.comment_SYNC import (
    Comment,
    LooseComment,
)

def load_comment(d: Mapping[str, Any], client: Client) -> Comment:
    return Comment(d, client)

def load_loose_comment(d: Mapping[str, Any], client: Client) -> LooseComment:
    return LooseComment(d, client)




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence
if TYPE_CHECKING:
    from .post_flair import PostFlairWidgetItem

from dataclasses import dataclass

from .widget_ASYNC import Widget

@dataclass(repr=False, eq=False)
class PostFlairWidget(Widget):
    display: str
    order: Sequence[str]
    templates: Sequence[PostFlairWidgetItem]




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, Optional, Mapping
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.subreddit_user import (
        Moderator,
        ApprovedUser,
        BannedUser,
        MutedUser,
    )
    from ...models.moderation_action_log_entry import ModerationActionLogEntry

from functools import cached_property

from ...model_loaders.subreddit_user import (
    load_moderator,
    load_approved_user,
    load_banned_user,
    load_muted_user,
)
from .pull_users_SYNC import PullUsers
from .legacy_SYNC import Legacy
from .pull_SYNC import Pull
from .note_SYNC import Note
from ...util.base_conversion import to_base36
from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.moderation.sync1 import ModerationActionLogPaginator

class ModerationProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.legacy: Legacy = Legacy(client)
        ("""
            Legacy procedures for retrieving subreddit users.
            """)
        self.pull_users: PullUsers = PullUsers(client)
        ("""
            Procedures for retrieving subreddit users.
            """)
        self.pull: Pull = Pull(client)
        ("""
            Procedures for pulling moderation items.
            """)
        self.note: Note = Note(client)
        ("""
            Procedures for managing mod notes.
            """)

    def pull_actions(self, sr: str, amount: Optional[int] = None, *,
            action: str = '', mod: str = '',
            ) -> ImpartedPaginatorChainingIterator[ModerationActionLogPaginator, ModerationActionLogEntry]:
        """Retrieve recent moderation actions from the mod log.

        .. hint::

           Moderator actions taken within a subreddit are logged to a mod log.
           Entries in the mod log last for 3 months before they become inaccessible.

        .. .PARAMETERS

        :param `str` sr:
            The target subreddit.
        :param `Optional[int]` amount:
            The number of items to retrieve.
        :param `str` action:
            Limit log entries to only those of a specified action.
        :param `str` mod:
            A comma-delimited list of moderator names to restrict the results to,
            or use the string `a` to restrict the results to actions performed by administrators.

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_iterator.ImpartedPaginatorChainingIterator`\\[:class:`~.pagination.paginators.moderation.sync1.ModerationActionLogPaginator`, :class:`~.models.moderation_action_log_entry.ModerationActionLogEntry`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                You do not have permission to view the mod log of the specified subreddit.
        """
        p = ModerationActionLogPaginator(self._client, f'/r/{sr}/about/log', action=action, mod=mod)
        return ImpartedPaginatorChainingIterator(p, amount)


    def get_moderator(self, sr: str, user: str) -> Optional[Moderator]:
        """
        Behaves similarly to :meth:`.get_approved_user`.
        """
        root = self._client.request('GET', f'/api/v1/{sr}/moderators', params={'username': user})
        order = root['moderatorIds']
        object_map = root['moderators']
        return load_moderator(object_map[order[0]]) if order else None

    def get_moderator_invitation(self, sr: str, user: str) -> Optional[Moderator]:
        """
        Behaves similarly to :meth:`.get_approved_user`.
        """
        root = self._client.request('GET', f'/api/v1/{sr}/moderators_invited', params={'username': user})
        order = root['moderatorIds']
        object_map = root['moderators']
        return load_moderator(object_map[order[0]]) if order else None

    def get_editable_moderator(self, sr: str, user: str) -> Optional[Moderator]:
        """
        Behaves similarly to :meth:`.get_approved_user`.
        """
        root = self._client.request('GET', f'/api/v1/{sr}/moderators_editable', params={'username': user})
        order = root['moderatorIds']
        object_map = root['moderators']
        return load_moderator(object_map[order[0]]) if order else None

    def get_approved_user(self, sr: str, user: str) -> Optional[ApprovedUser]:
        """Get an approved user of a subreddit.

        .. .PARAMETERS

        :param `str` sr:
            Name of a subreddit.
        :param `str` user:
            Name of a user.

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.subreddit_user.ApprovedUser`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `SUBREDDIT_NOEXIST`:
                The target subreddit does not exist.
            + `SUBREDDIT_NO_ACCESS`:
                The subreddit cannot be accessed.
        """
        root = self._client.request('GET', f'/api/v1/{sr}/contributors', params={'username': user})
        order = root['approvedSubmitterIds']
        object_map = root['approvedSubmitters']
        return load_approved_user(object_map[order[0]]) if order else None

    def get_banned_user(self, sr: str, user: str) -> Optional[BannedUser]:
        """
        Behaves similarly to :meth:`.get_approved_user`.
        """
        root = self._client.request('GET', f'/api/v1/{sr}/banned', params={'username': user})
        order = root['bannedUserIds']
        object_map = root['bannedUsers']
        return load_banned_user(object_map[order[0]]) if order else None

    def get_muted_user(self, sr: str, user: str) -> Optional[MutedUser]:
        """
        Behaves similarly to :meth:`.get_approved_user`.
        """
        root = self._client.request('GET', f'/api/v1/{sr}/muted', params={'username': user})
        order = root['mutedUserIds']
        object_map = root['mutedUsers']
        return load_muted_user(object_map[order[0]]) if order else None


    def send_moderator_invite(self, sr: str, user: str, permissions: Sequence[str]) -> None:
        """Send a moderator invite.

        If the user is already invited, nothing happens. The permissions won't change.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `Sequence[str]` permissions:
            Values: `all`, `access`, `chat_config`, `chat_operator`, `config`,
            `flair`, `mail`, `posts`, `wiki`.

            Pass an empty sequence for no permissions.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The `user` parameter was empty, or contains invalid characters.
            + `USER_DOESNT_EXIST`:
                The user specified by `user` does not exist.
            + `INVALID_PERMISSIONS`:
                An invalid permission was provided.
            + `ALREADY_MODERATOR`:
                The user is already a moderator.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - You don't have permission.
               - The target subreddit specified was empty.

            + `404`:
                The specified subreddit does not exist.
        """
        data = {
            'type': 'moderator_invite',
            'r': sr,
            'name': user,
            'permissions': ','.join('+' + p for p in permissions) if permissions else '-access',
        }
        self._client.request('POST', '/api/friend', data=data)

    def accept_moderator_invite(self, sr: str) -> None:
        """Accept a moderator invite.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_INVITE_FOUND`:
                You don't have a pendning invitation for the specified subreddit.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The subreddit specified was empty.
        """
        self._client.request('POST', '/api/accept_moderator_invite', data={'r': sr})

    def revoke_moderator_invite(self, sr: str, user: str) -> None:
        """Revoke a moderator invite.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `400`:
                The `user` parameter was empty or the user doesn't exist.
            + `403`:
               - You don't have permission.
               - The target subreddit specified was empty.
        """
        data = {
            'type': 'moderator_invite',
            'r': sr,
            'name': user,
        }
        self._client.request('POST', '/api/unfriend', data=data)

    def remove_moderator(self, sr: str, user: str) -> None:
        """Remove a moderator.

        Nothing happens if the specified user is not a moderator of the subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `400`:
                The `user` parameter was empty or the user doesn't exist.
            + `403`:
               - You don't have permission.
               - The target subreddit specified was empty.
        """
        data = {
            'type': 'moderator',
            'r': sr,
            'name': user,
        }
        self._client.request('POST', '/api/unfriend', data=data)

    def leave_moderator(self, sr_id: int) -> None:
        """Abdicate moderator status in a subreddit.

        .. warning::

            This action cannot be undone.

        Be careful with this endpoint. It's possible for a subreddit to not have any moderators.

        Nothing happens if the specified subreddit ID is not valid or the
        current user is not a moderator of the subreddit.

        .. .PARAMETERS

        :param `int` sr_id:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        self._client.request('POST', '/api/leavemoderator', data={'id': 't5_' + to_base36(sr_id)})

    def set_moderator_permissions(self, sr: str, user: str, permissions: Sequence[str]) -> None:
        """Set the permissions of a moderator.

        Be careful with this endpoint. It's possible for a subreddit to not have any moderators.

        Nothing happens if the current user is not a moderator of the subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `Sequence[str]` permissions:
            Values: `all`, `access`, `chat_config`, `chat_operator`, `config`,
            `flair`, `mail`, `posts`, `wiki`.

            Pass an empty sequence for no permissions.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The `user` parameter was empty, or contains invalid characters.
            + `USER_DOESNT_EXIST`:
                The user specified by `user` does not exist.
            + `INVALID_PERMISSION_TYPE`:
                The user specified by `user` isn't a moderator of the subreddit.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You don't have permission.
            + `404`:
                The specified subreddit does not exist.
        """
        data = {
            'type': 'moderator',
            'r': sr,
            'name': user,
            'permissions': ','.join('+' + p for p in permissions) if permissions else '-access',
        }
        self._client.request('POST', '/api/setpermissions', data=data)

    def set_moderator_invite_permissions(self, sr: str, user: str, permissions: Sequence[str]) -> None:
        """Set the permissions on a moderator invite.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `Sequence[str]` permissions:
            Values: `all`, `access`, `chat_config`, `chat_operator`, `config`,
            `flair`, `mail`, `posts`, `wiki`.

            Pass an empty sequence for no permissions.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The `user` parameter was empty, or contains invalid characters.
            + `USER_DOESNT_EXIST`:
                The user specified by `user` does not exist.
            + `INVALID_PERMISSION_TYPE`:
                The user specified by `user` doesn't have a moderator invite to the subreddit.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You don't have permission.
            + `404`:
                The specified subreddit does not exist.
        """
        data = {
            'type': 'moderator_invite',
            'r': sr,
            'name': user,
            'permissions': ','.join('+' + p for p in permissions) if permissions else '-access',
        }
        self._client.request('POST', '/api/setpermissions', data=data)

    def add_approved_user(self, sr: str, user: str) -> None:
        """Add an approved user to a subreddit.

        If the user is already an approved user, nothing happens.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The `user` parameter was empty, or contains invalid characters.
            + `USER_DOESNT_EXIST`:
                The user specified by `user` does not exist.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - You don't have permission.
               - The target subreddit specified was empty.

            + `404`:
                The specified subreddit does not exist.
        """
        data = {
            'type': 'contributor',
            'r': sr,
            'name': user,
        }
        self._client.request('POST', '/api/friend', data=data)

    def remove_approved_user(self, sr: str, user: str) -> None:
        """Remove an approved user of a subreddit.

        If the user is already not an approved user, nothing happens.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The `user` parameter was empty, or contains invalid characters.
            + `USER_DOESNT_EXIST`:
                The user specified by `user` does not exist.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `400`:
                The `user` parameter was empty or the user doesn't exist.
            + `403`:
               - You don't have permission.
               - The target subreddit specified was empty.

            + `404`:
                The specified subreddit does not exist.
        """
        data = {
            'type': 'contributor',
            'r': sr,
            'name': user,
        }
        self._client.request('POST', '/api/unfriend', data=data)

    def leave_approved_user(self, sr_id: int) -> None:
        """Abdicate approved user status in a subreddit.

        Nothing happens if the specified subreddit ID is not valid or the
        current user is not an approved user of the subreddit.

        .. .PARAMETERS

        :param `int` sr_id:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        self._client.request('POST', '/api/leavecontributor', data={'id': 't5_' + to_base36(sr_id)})

    def ban_user(self, sr: str, user: str, *,
            duration: Optional[int] = None,
            reason: str = '',
            note: str = '',
            message: str = '') -> None:
        """Ban a user from a subreddit.

        Banning an already banned user does nothing. However, when banning an already
        banned user (as to change the ban reason or note), if the duration is changed
        then a new direct message will be issued to the user informing them of the
        duration change.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `Optional[int]` duration:
            Number of days they should be banned for, in the range of 1 to 999.

            To change the duration of a ban, re-issue this procedure with a new duration.
            A direct message is sent to the user informing them of the ban duration change.
        :param `str` reason:
            Ban reason. No longer than 100 characters.
        :param `str` note:
            A moderator note. No longer than 300 characters.
        :param `str` message:
            The message to include in the DM that is sent to the user.

            Specify markdown text.

            Note that a direct message is always sent to the banned user when a ban is issued.
            The ban message shows in the DM under a section called "Note from the moderators:".

            If empty string, no message or section "Note from the moderators:" is shown.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The `user` parameter was empty, or contains invalid characters.
            + `USER_DOESNT_EXIST`:
                The user specified by `user` does not exist.
            + `BAD_NUMBER`:
                When `type: banned` or `type: wikibanned`,
                the number specified by `duration` is not within the range 1 to 999.
            + `TOO_LONG`:
               - The value specified by `reason` is over 100 characters.
               - The value specified by `note` is over 300 characters.

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - You don't have permission.
               - The target subreddit specified was empty.

            + `404`:
                The specified subreddit does not exist.
        """
        data = {
            'type': 'banned',
            'r': sr,
            'name': user,
            'ban_reason': reason,
            'ban_message': message,
            'note': note,
        }
        if duration is not None:
            data['duration'] = str(duration)

        self._client.request('POST', '/api/friend', data=data)

    def unban_user(self, sr: str, user: str) -> None:
        """Unban a user from a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises): Same as :meth:`~.remove_approved_user`.
        """
        data = {
            'type': 'banned',
            'r': sr,
            'name': user,
        }
        self._client.request('POST', '/api/unfriend', data=data)

    def mute_user(self, sr: str, user: str, *, note: str = '') -> None:
        """Mute a user in a subreddit.

        Muting an already muted user does nothing.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `str` note:
            A moderator note. No longer than 300 characters.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            In addition to :meth:`~.add_approved_user`:

            + `TOO_LONG`:
                The value specified by `note` is over 300 characters.
        :raises redditwarp.http.exceptions.StatusCodeException:
            Same as :meth:`~.add_approved_user`.
        """
        data = {
            'type': 'muted',
            'r': sr,
            'name': user,
            'note': note,
        }
        self._client.request('POST', '/api/friend', data=data)

    def unmute_user(self, sr: str, user: str) -> None:
        """Unmute a user in a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises): Same as :meth:`~.remove_approved_user`.
        """
        data = {
            'type': 'muted',
            'r': sr,
            'name': user,
        }
        self._client.request('POST', '/api/unfriend', data=data)

    def add_wiki_contributor(self, sr: str, user: str) -> None:
        """Add a wiki contributor in a subreddit.

        Adding a user who is already a wiki contributor does nothing.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises): Same as :meth:`~.add_approved_user`.
        """
        data = {
            'type': 'wikicontributor',
            'r': sr,
            'name': user,
        }
        self._client.request('POST', '/api/friend', data=data)

    def remove_wiki_contributor(self, sr: str, user: str) -> None:
        """Remove a wiki contributor in a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises): Same as :meth:`~.remove_approved_user`.
        """
        data = {
            'type': 'wikicontributor',
            'r': sr,
            'name': user,
        }
        self._client.request('POST', '/api/unfriend', data=data)

    def ban_wiki_contributor(self, sr: str, user: str, *,
            duration: Optional[int] = None,
            reason: str = '',
            note: str = '') -> None:
        """Ban a wiki contributor in a subreddit.

        Adding a user who is already a wiki contributor does nothing.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `Optional[int]` duration:
            Number of days they should be banned for, in the range of 1 to 999.
        :param `str` reason:
            Ban reason. No longer than 100 characters.
        :param `str` note:
            A moderator note. No longer than 300 characters.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            In addition to :meth:`~.add_approved_user`:

            + `TOO_LONG`:
               - The value specified by `reason` is over 100 characters.
               - The value specified by `note` is over 300 characters.

        :raises redditwarp.http.exceptions.StatusCodeException:
            Same as :meth:`~.add_approved_user`.
        """
        data = {
            'type': 'banned',
            'r': sr,
            'name': user,
            'ban_reason': reason,
            'note': note,
        }
        if duration is not None:
            data['duration'] = str(duration)

        self._client.request('POST', '/api/friend', data=data)

    def unban_wiki_contributor(self, sr: str, user: str) -> None:
        """Unban a wiki contributor in a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises): Same as :meth:`~.remove_approved_user`.
        """
        data = {
            'type': 'wikibanned',
            'r': sr,
            'name': user,
        }
        self._client.request('POST', '/api/unfriend', data=data)


    class RemovalReason:
        def __init__(self, outer: ModerationProcedures) -> None:
            self._outer = outer
            self._client = outer._client

        def create(self, sr: str, title: str, message: str) -> str:
            """Create a removal reason.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` title:
                A title for this removal reason.
            :param `str` message:
                The removal reason message.

            .. .RETURNS

            :returns: The newly created removal reason ID.
            :rtype: `str`

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `MOD_REQUIRED`:
                   - There is no user context.
                   - The current user is not a moderator of the subreddit.

                + `NO_TEXT`:
                   - The `title` parameter was not specified or was empty.
                   - The `message` parameter was not specified or was empty.

                + `TOO_LONG`:
                   - The value specified for `title` is over 50 characters.
                   - The value specified for `message` is over 10000 characters.

            :raises redditwarp.http.exceptions.StatusCodeException:
                + `500`:
                    The target subreddit does not exist.
            """
            data = {'title': title, 'message': message}
            root = self._client.request('POST', f'/api/v1/{sr}/removal_reasons', data=data)
            return root['id']

        def retrieve(self, sr: str) -> Mapping[str, tuple[str, str]]:
            """Get a list of removal reasons.

            .. .PARAMETERS

            :param `str` sr:

            .. .RETURNS

            :returns: A mapping of removal reason IDs to tuples of `(title, message)`.
            :rtype: `Mapping`\\[`str`, `tuple`\\[`str`, `str`]]

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `MOD_REQUIRED`:
                   - There is no user context.
                   - The current user is not a moderator of the subreddit.

                + `SUBREDDIT_NOEXIST`:
                    The target subreddit does not exist.
            """
            root = self._client.request('GET', f'/api/v1/{sr}/removal_reasons')
            order = root['order']
            object_map = root['data']
            return {
                y: (m['title'], m['message'])
                for y in order for m in [object_map[y]]
            }

        def replace(self, sr: str, idt: str, title: str, message: str) -> None:
            """Update a removal reason's title and message.

            Both parameters `title` and `message` must be specified otherwise
            a `NO_TEXT` API error is returned.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` idt:
                A removal reason ID.
            :param `str` title:
            :param `str` message:

            .. .RETURNS

            :rtype: `None`

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `MOD_REQUIRED`:
                   - There is no user context.
                   - The current user is not a moderator of the subreddit.

                + `SUBREDDIT_NOEXIST`:
                    The target subreddit does not exist.
                + `INVALID_ID`:
                    The specified removal reason ID was not found.
                + `NO_TEXT`:
                   - The `title` parameter was not specified or was empty.
                   - The `message` parameter was not specified or was empty.

                + `TOO_LONG`:
                   - The value specified for `title` is over 50 characters.
                   - The value specified for `message` is over 10000 characters.
            """
            data = {'title': title, 'message': message}
            self._client.request('PUT', f'/api/v1/{sr}/removal_reasons/{idt}', data=data)

        def delete(self, sr: str, idt: str) -> None:
            """Delete a removal reason.

            If the specified removal reason ID did not exist, nothing happens.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` idt:

            .. .RETURNS

            :rtype: `None`

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `MOD_REQUIRED`:
                   - There is no user context.
                   - The current user is not a moderator of the subreddit.

                + `SUBREDDIT_NOEXIST`:
                    The target subreddit does not exist.
                + `INVALID_ID`:
                    The specified removal reason ID was not found.
            """
            self._client.request('DELETE', f'/api/v1/{sr}/removal_reasons/{idt}')

    removal_reason: cached_property[RemovalReason] = cached_property(RemovalReason)
    ("""
        Manage subreddit removal reasons.
        """)




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable
if TYPE_CHECKING:
    from ...client_SYNC import Client

from .listing.mixins.sort_SYNC import Sort
from .listing.subreddit_listing_paginator import SubredditListingPaginator
from ...models.subreddit_SYNC import Subreddit

class SubredditSearchPaginator(
    Sort[Subreddit],
    SubredditListingPaginator,
):
    def __init__(self,
        client: Client,
        url: str,
        query: str,
        show_users: bool = False,
    ) -> None:
        super().__init__(client, url)
        self.query: str = query
        ("")
        self.show_users: bool = show_users
        ("")

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        yield ('q', self.query)
        if self.show_users:
            yield ('show_users', '1')




from __future__ import annotations
from typing import Mapping, Any, Sequence, Iterator, Union, overload, Optional

from datetime import datetime, timezone

class LiveThread:
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.idt: str = d['id']
        ("""
            E.g., `177beztuzebxj`.
            """)
        self.title: str = d['title']
        ("")
        self.created_ut: int = int(d['created_utc'])
        ("")
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("")
        self.description: str = d['description']
        ("""
            Live thread description as markdown text.

            This is the text below the title.
            """)
        self.description_html: str = d['description_html']
        ("""
            Live thread description as HTML.
            """)
        self.resources: str = d['resources']
        ("""
            Sidebar text in markdown.
            """)
        self.resources_html: str = d['resources_html']
        ("""
            Sidebar text in HTML.
            """)
        self.websocket_url: str = d['websocket_url']
        ("""
            Websocket URL. Connect to this websocket to stream live updates.
            """)
        self.complete: bool = d['state'] == 'complete'
        ("""
            True if the live thread is marked complete.
            """)
        self.nsfw: bool = d['nsfw']
        ("")
        self.viewer_count: int = d['viewer_count']
        ("""
            Number of subscribers. Value is fuzzed
            """)

class LiveUpdate:
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.uuid: str = d['id']
        ("""
            E.g., `890e9242-d7fb-11eb-b450-0ed185f1b209`.
            """)
        self.author: Optional[str] = d['author']
        ("""
            Name of the user who posted the update.

            Value null if the user account was deleted.
            """)
        self.body: str = d['body']
        ("""
            Markdown content body.
            """)
        self.body_html: str = d['body_html']
        ("""
            Content body in HTML.
            """)
        self.created_ut: int = int(d['created_utc'])
        ("")
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("")
        self.stricken: bool = d['stricken']
        ("""
            True if the update has been stricken.
            """)

class Contributor:
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.name: str = d['name']
        ("""
            Name of a user; name of the contributor.
            """)
        full_id36: str = d['id']
        _, _, id36 = full_id36.partition('_')
        self.id36: str = id36
        ("""
            Contributor's user ID as a base 36 number.
            """)
        self.id: int = int(id36, 36)
        ("""
            Contributor's user ID.
            """)
        self.permissions: Sequence[str] = d['permissions']
        ("""
            Values: `all`, `close`, `discussions`, `edit`, `manage`, `settings`, `update`.

            Can be empty. This means no permissions.
            """)

class ContributorList(Sequence[Contributor]):
    def __init__(self,
            contributors: Sequence[Contributor],
            invitations: Sequence[Contributor]) -> None:
        self.contributors: Sequence[Contributor] = contributors
        ("")
        self.invitations: Sequence[Contributor] = invitations
        ("")

    def __len__(self) -> int:
        return len(self.contributors)

    def __contains__(self, item: object) -> bool:
        return item in self.contributors

    def __iter__(self) -> Iterator[Contributor]:
        return iter(self.contributors)

    @overload
    def __getitem__(self, index: int) -> Contributor: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[Contributor]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[Contributor, Sequence[Contributor]]:
        return self.contributors[index]




from __future__ import annotations
from typing import TypeVar, Callable, Generic

TInput = TypeVar('TInput')
TOutput = TypeVar('TOutput')

class CallChunk(Generic[TInput, TOutput]):
    """Perform `.operation` on `.operand` when called."""

    def __init__(self,
        operation: Callable[[TInput], TOutput],
        operand: TInput,
    ) -> None:
        self.operation: Callable[[TInput], TOutput] = operation
        ("")
        self.operand: TInput = operand
        ("")

    def __call__(self) -> TOutput:
        return self.operation(self.operand)




from __future__ import annotations

from .listing.mixins.subreddit_detail_ASYNC import SubredditDetail
from .listing.submission_listing_async_paginator import SubmissionListingAsyncPaginator
from ...models.submission_ASYNC import Submission
from .listing.mixins.time_ASYNC import Time
from .listing.submission_and_comment_listing_async_paginator import SubmissionAndLooseCommentListingAsyncPaginator

class BestListingAsyncPaginator(
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class HotListingAsyncPaginator(
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class NewListingAsyncPaginator(
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class TopListingAsyncPaginator(
    Time[Submission],
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class RisingListingAsyncPaginator(
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class ControversialListingAsyncPaginator(
    Time[Submission],
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class GildedListingAsyncPaginator(
    Time[object],
    SubredditDetail[object],
    SubmissionAndLooseCommentListingAsyncPaginator,
): pass




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..http.handler_ASYNC import Handler
    from ..http.send_params import SendParams
    from ..http.exchange import Exchange

from ..util.imports import lazy_import
if TYPE_CHECKING:
    import asyncio
else:
    lazy_import % 'asyncio'

import time
from datetime import datetime, timezone
from email.utils import parsedate_to_datetime

from ..http.delegating_handler_ASYNC import DelegatingHandler
from ..util.token_bucket import TokenBucket

class RateLimited(DelegatingHandler):
    def __init__(self, handler: Handler) -> None:
        super().__init__(handler)
        self.remaining: int = 0
        ("")
        self.reset: int = 0
        ("")
        self.used: int = 0
        ("")
        self._delta: float = 0.
        self._timestamp: float = time.monotonic()
        self._tb = TokenBucket(10, 1)
        self._lock = asyncio.Lock()
        self._datetime_extremum = datetime.min.replace(tzinfo=timezone.utc)

    async def _send(self, p: SendParams) -> Exchange:
        tb = self._tb
        async with self._lock:
            s = 0.
            if self.remaining <= 1:
                s = self.reset
            elif (w := self.reset / self.remaining) >= 2:
                # If the API wants us to wait for longer than two seconds then oblige.
                s = w

            if s:
                # Don't add tokens for the time spent sleeping here.
                bv = tb.get_value()
                await asyncio.sleep(s)
                tb.consume(tb.get_value() - bv)

            # The token bucket rate limiting is done in conjunction to the other rate limiting logic.
            await asyncio.sleep(tb.get_cooldown(1))
            tb.consume(1)

        xchg = await super()._send(p)

        now = time.monotonic()
        self._delta = now - self._timestamp
        self._timestamp = now

        date_string = xchg.response.headers['Date']
        dt = parsedate_to_datetime(date_string)
        if dt > self._datetime_extremum:
            self._datetime_extremum = dt

            headers = xchg.response.headers
            if 'x-ratelimit-reset' in headers:
                self.remaining = int(float(headers['x-ratelimit-remaining']))
                self.reset = int(headers['x-ratelimit-reset'])
                self.used = int(headers['x-ratelimit-used'])
            else:
                if self.reset > 0:
                    self.remaining -= 1
                    self.reset = max(0, self.reset - int(self._delta))
                    self.used += 1
                else:
                    self.remaining = 300
                    self.reset = 600
                    self.used = 0

        return xchg




from . import handler_ASYNC  # noqa: F401
from .handler_ASYNC import Handler as Handler  # noqa: F401
from . import delegating_handler_ASYNC  # noqa: F401
from .delegating_handler_ASYNC import DelegatingHandler as DelegatingHandler  # noqa: F401
from . import http_client_ASYNC  # noqa: F401
from .http_client_ASYNC import HTTPClient as HTTPClient  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.submission_ASYNC import Submission
    from ...models.comment_ASYNC import Comment

from functools import cached_property

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.moderation.async1 import (
    ModQueueListingAsyncPaginator,
    ModQueueSubmissionListingAsyncPaginator,
    ModQueueCommentListingAsyncPaginator,
    ReportsListingAsyncPaginator,
    ReportsSubmissionListingAsyncPaginator,
    ReportsCommentListingAsyncPaginator,
    SpamListingAsyncPaginator,
    SpamSubmissionListingAsyncPaginator,
    SpamCommentListingAsyncPaginator,
    EditedListingAsyncPaginator,
    EditedSubmissionListingAsyncPaginator,
    EditedCommentListingAsyncPaginator,
    UnmoderatedSubmissionListingAsyncPaginator,
)

class Pull:
    def __init__(self, client: Client) -> None:
        self._client = client

    class ModQueue:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ModQueueListingAsyncPaginator, object]:
            p = ModQueueListingAsyncPaginator(self._client, f'/r/{sr}/about/modqueue')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ModQueueSubmissionListingAsyncPaginator, Submission]:
            p = ModQueueSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/modqueue')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ModQueueCommentListingAsyncPaginator, Comment]:
            p = ModQueueCommentListingAsyncPaginator(self._client, f'/r/{sr}/about/modqueue')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

    modqueue: cached_property[ModQueue] = cached_property(ModQueue)
    ("""
        Retrieve submissions/comments relevant to moderators.

        .. .PARAMETERS

        :param `str` sr:
        :param `Optional[int]` amount:

        .. .RETURNS

        :returns:
            A paginator iterator of either :class:`~.models.submission_ASYNC.Submission`
            or :class:`~.models.comment_ASYNC.Comment` objects.
        :rtype: :class:`~.pagination.paginator_chaining_async_iterator.ImpartedPaginatorChainingAsyncIterator`\\[:class:`~.pagination.paginators.moderation.async1.ModQueueListingAsyncPaginator`, `object`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - You don't have access to the subreddit.
               - You don't have the 'posts' moderator permission.
            + `403`:
                The specified subreddit name is too long (over 21 characters)
                or contains invalid characters.
        """)

    class Reported:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ReportsListingAsyncPaginator, object]:
            p = ReportsListingAsyncPaginator(self._client, f'/r/{sr}/about/reports')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ReportsSubmissionListingAsyncPaginator, Submission]:
            p = ReportsSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/reports')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ReportsCommentListingAsyncPaginator, Comment]:
            p = ReportsCommentListingAsyncPaginator(self._client, f'/r/{sr}/about/reports')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

    reported: cached_property[Reported] = cached_property(Reported)
    ("""
        Behaves similarly to :meth:`.modqueue`.
        """)

    class Spam:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[SpamListingAsyncPaginator, object]:
            p = SpamListingAsyncPaginator(self._client, f'/r/{sr}/about/spam')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[SpamSubmissionListingAsyncPaginator, Submission]:
            p = SpamSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/spam')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[SpamCommentListingAsyncPaginator, Comment]:
            p = SpamCommentListingAsyncPaginator(self._client, f'/r/{sr}/about/spam')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

    spam: cached_property[Spam] = cached_property(Spam)
    ("""
        Behaves similarly to :meth:`.modqueue`.
        """)

    class Edited:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[EditedListingAsyncPaginator, object]:
            p = EditedListingAsyncPaginator(self._client, f'/r/{sr}/about/edited')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[EditedSubmissionListingAsyncPaginator, Submission]:
            p = EditedSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/edited')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[EditedCommentListingAsyncPaginator, Comment]:
            p = EditedCommentListingAsyncPaginator(self._client, f'/r/{sr}/about/edited')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

    edited: cached_property[Edited] = cached_property(Edited)
    ("""
        Behaves similarly to :meth:`.modqueue`.
        """)

    def unmoderated(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[UnmoderatedSubmissionListingAsyncPaginator, Submission]:
        """
        Behaves similarly to :meth:`.modqueue`.
        """
        p = UnmoderatedSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/unmoderated')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .send_params import SendParams
    from .exchange import Exchange

from .delegating_handler_SYNC import DelegatingHandler


class Invoker(DelegatingHandler):
    def send(self, p: SendParams) -> Exchange:
        return self._send(p)

    def close(self) -> None:
        self._close()




from __future__ import annotations
from typing import Sequence, overload, Iterator, Union, Mapping, Any, Optional

from datetime import datetime
from dataclasses import dataclass

from .datamemento import DatamementoDataclassesMixin
from .subreddit import Subreddit

@dataclass(repr=False, eq=False)
class SubmissionDraft(DatamementoDataclassesMixin):
    """
    A public draft link is of the following format:

    `https://www.reddit.com/user/{USERNAME}/draft/{self.uuid}`
    """
    @dataclass(repr=False, eq=False)
    class Flair:
        uuid: str
        ("""
            The chosen flair template UUID.
            """)
        text_mode: str
        ("""
            Either `text` or `richtext`.
            """)
        text: str
        ("""
            Flair text.
            """)
        bg_color: str
        ("""
            Reminder: cannot be an empty string since this is isn't a user flair.
            """)
        fg_color_scheme: str
        ("""
            Either `dark` or `light`.
            """)

    d: Mapping[str, Any]
    uuid: str
    created_at: datetime
    ("""
        Datetime object of when the draft was created.
        """)
    modified_at: datetime
    ("""
        Datetime object of when the draft was last modified.
        """)
    public: bool
    ("""
        Whether the draft's link is public.

        Only those with the link can find the draft.
        """)
    subreddit_id: Optional[int]
    ("""
        The ID36 of the target subreddit.

        Value is null if not chosen yet.
        """)
    title: str
    reply_notifications: bool
    spoiler: bool
    nsfw: bool
    oc: bool
    ("""
        Whether the post should be marked as 'original content'.
        """)
    flair: Optional[Flair]
    ("""
        A flair option from the target subreddit.

        Value `None` if no flair selected.
        """)


class TextPostDraft(SubmissionDraft):
    pass

@dataclass(repr=False, eq=False)
class MarkdownTextPostDraft(TextPostDraft):
    body: str
    ("""
        The body text of the submission draft. In markdown format.
        """)

class RichTextTextPostDraft(TextPostDraft):
    pass


@dataclass(repr=False, eq=False)
class LinkPostDraft(SubmissionDraft):
    link: str
    ("""
        The linked URL.
        """)



class SubmissionDraftList(Sequence[SubmissionDraft]):
    @property
    def subreddits(self) -> Sequence[Subreddit]:
        return self.__subreddits

    def __init__(self,
        drafts: Sequence[SubmissionDraft],
        subreddits: Sequence[Subreddit],
    ) -> None:
        self.drafts: Sequence[SubmissionDraft] = drafts
        ("")
        self.__subreddits: Sequence[Subreddit] = subreddits

    def __len__(self) -> int:
        return len(self.drafts)

    def __contains__(self, item: object) -> bool:
        return item in self.drafts

    def __iter__(self) -> Iterator[SubmissionDraft]:
        return iter(self.drafts)

    @overload
    def __getitem__(self, index: int) -> SubmissionDraft: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[SubmissionDraft]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[SubmissionDraft, Sequence[SubmissionDraft]]:
        return self.drafts[index]




from __future__ import annotations
from typing import Optional

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class HoverState:
    pass

@dataclass(repr=False, eq=False)
class TextHoverState(HoverState):
    label: str
    text_color: str
    fill_color: str
    stroke_color: str

@dataclass(repr=False, eq=False)
class ImageHoverState(HoverState):
    image_url: str
    image_size: tuple[int, int]


@dataclass(repr=False, eq=False)
class Button:
    label: str
    link: str
    hover_state: Optional[HoverState]

@dataclass(repr=False, eq=False)
class TextButton(Button):
    text_color: str
    fill_color: str
    stroke_color: str

@dataclass(repr=False, eq=False)
class ImageButton(Button):
    image_url: str
    image_size: tuple[int, int]




from typing import Union, Mapping, Sequence, Dict, List

# https://github.com/python/typing/issues/182#issuecomment-1320974824
JSON = Union[Dict[str, 'JSON'], List['JSON'], str, int, float, bool, None]
JSON_ro = Union[Mapping[str, 'JSON_ro'], Sequence['JSON_ro'], str, int, float, bool, None]




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional, Sequence, Iterable
if TYPE_CHECKING:
    from ...client_ASYNC import Client

from ..async_paginator import HasMoreAsyncPaginator, Bidirectional, CursorAsyncPaginator
from ...models.flair import UserFlairAssociation
from ...model_loaders.flair import load_user_flair_association

class UserFlairAssociationAsyncPaginator(
        HasMoreAsyncPaginator[UserFlairAssociation], Bidirectional, CursorAsyncPaginator[UserFlairAssociation]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        limit: Optional[int] = 1000,
    ) -> None:
        super().__init__()
        self.limit: Optional[int] = limit
        ("")
        self.direction: bool = True
        ("")
        self.client: Client = client
        ("")
        self.url: str = url
        ("")
        self._after: str = ''
        self._before: str = ''
        self._has_after: bool = True
        self._has_before: bool = True

    def get_cursor(self) -> str:
        return self._after if self.direction else self._before

    def set_cursor(self, value: str) -> None:
        if self.direction:
            self._after = value
        else:
            self._before = value

    def has_more(self) -> bool:
        return self._has_after if self.direction else self._has_before

    def set_has_more(self, value: bool) -> None:
        if self.direction:
            self._has_after = value
        else:
            self._has_before = value

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        if self.limit is not None:
            yield ('limit', str(self.limit))

        if self.direction:
            if self._after:
                yield ('after', self._after)
        else:
            if self._before:
                yield ('before', self._before)

    async def _fetch_data(self) -> Any:
        params = dict(self._generate_params())
        data = await self.client.request('GET', self.url, params=params)
        children = data['users']

        suggested_forward_cursor = data.get('after', '')
        suggested_backward_cursor = data.get('before', '')
        if children:
            self._after = suggested_forward_cursor
            self._before = suggested_backward_cursor
        self._has_after = bool(suggested_forward_cursor)
        self._has_before = bool(suggested_backward_cursor)

        return data

    async def fetch(self) -> Sequence[UserFlairAssociation]:
        data = await self._fetch_data()
        return [load_user_flair_association(d) for d in data['users']]




from __future__ import annotations
from typing import Optional, Iterable

import time
import json
import re

from ....core.http_client_SYNC import HTTPClient
from ....http.util.json_loading import load_json_from_response_but_prefer_status_code_exception_on_failure
from ....exceptions import (
    OperationException,
    raise_for_reddit_error,
)
from ....util.redditwarp_installed_client_credentials import get_device_id
from ...util.request_signing import (
    SIGNATURE_UA_MESSAGE_TEMPLATE,
    SIGNATURE_BODY_MESSAGE_TEMPLATE,
    generate_android_mobile_hmac_hash,
    get_android_mobile_request_signature,
)


class LoginProcedures:
    def __init__(self, http: HTTPClient) -> None:
        self._http = http

    def do_legacy_web_login(self, username: str, password: str, otp: Optional[str] = None) -> tuple[str, str]:
        def g() -> Iterable[tuple[str, str]]:
            yield ('user', username)
            yield ('passwd', password)
            if otp is not None: yield ('otp', otp)

        resp = self._http.request(
                'POST', 'https://old.reddit.com/api/login',
                params={'api_type': 'json'},
                data=dict(g()))

        json_data = load_json_from_response_but_prefer_status_code_exception_on_failure(resp)

        raise_for_reddit_error(json_data)

        d = json_data['json']['data']
        if d.get('details', '') == 'TWO_FA_REQUIRED':
            raise OperationException('TWO_FA_REQUIRED')

        resp.ensure_successful_status()

        return (d['cookie'], d['modhash'])

    def do_modern_web_login(self, username: str, password: str, otp: Optional[str] = None) -> None:
        resp = self._http.request('GET', 'https://www.reddit.com/login/')
        resp.ensure_successful_status()

        body = resp.data.decode()
        m = re.search(r'''<input type="hidden" name="csrf_token" value="(\w+)">''', body)
        if m is None:
            raise RuntimeError
        csrf_token = m[1]

        def g() -> Iterable[tuple[str, str]]:
            yield ('csrf_token', csrf_token)
            yield ('username', username)
            yield ('password', password)
            if otp is not None: yield ('otp', otp)

        resp = self._http.request('POST', 'https://www.reddit.com/login', data=dict(g()))

        json_data = load_json_from_response_but_prefer_status_code_exception_on_failure(resp)

        raise_for_reddit_error(json_data)

        if json_data.get('details', '') == 'TWO_FA_REQUIRED':
            raise OperationException('TWO_FA_REQUIRED')

        resp.ensure_successful_status()

    def do_android_mobile_login(self, username: str, password: str, otp: Optional[str] = None) -> tuple[int, str]:
        epoch = int(time.time())
        ua = self._http.get_user_agent()
        vendor = get_device_id()

        msg = SIGNATURE_UA_MESSAGE_TEMPLATE % (epoch, ua, vendor)
        hsh = generate_android_mobile_hmac_hash(msg)
        ua_sig = get_android_mobile_request_signature(epoch, hsh)

        body_obj = {
            "username": username,
            "password": password,
            **({} if otp is None else {"otp": otp}),
        }
        body = json.dumps(body_obj, separators=(',', ':'))
        msg = SIGNATURE_BODY_MESSAGE_TEMPLATE % (epoch, body)
        hsh = generate_android_mobile_hmac_hash(msg)
        body_sig = get_android_mobile_request_signature(epoch, hsh)

        resp = self._http.request(
                'POST', 'https://accounts.reddit.com/api/login',
                headers={
                    'User-Agent': ua,
                    'client-vendor-id': vendor,
                    'X-hmac-signed-result': ua_sig,
                    'X-hmac-signed-body': body_sig,
                },
                data=body.encode())

        json_data = load_json_from_response_but_prefer_status_code_exception_on_failure(resp)

        raise_for_reddit_error(json_data.get('error'))

        resp.ensure_successful_status()

        _, _, user_id36 = json_data['userId'].partition('_')
        return (int(user_id36, 36), json_data['modhash'])




from __future__ import annotations

from dataclasses import dataclass

from .widget_SYNC import Widget

@dataclass(repr=False, eq=False)
class TextAreaWidget(Widget):
    text: str
    text_html: str




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Mapping, MutableMapping, Union
if TYPE_CHECKING:
    from .payload import Payload
    from .types import RequestFiles
    from ..types import JSON_ro

from dataclasses import dataclass

from .payload import make_payload
from .util.case_insensitive_dict import CaseInsensitiveDict


@dataclass(repr=False, eq=False)
class Requisition:
    """Basically a request object designed to be mutated."""
    verb: str
    url: str
    params: MutableMapping[str, str]
    headers: MutableMapping[str, str]
    payload: Optional[Payload]


def make_requisition(
    verb: str,
    url: str,
    *,
    params: Optional[Mapping[str, str]] = None,
    headers: Optional[Mapping[str, str]] = None,
    data: Optional[Union[Mapping[str, str], bytes]] = None,
    json: JSON_ro = None,
    files: Optional[RequestFiles] = None,
    payload: Optional[Payload] = None,
) -> Requisition:
    params = dict(params or {})
    headers = CaseInsensitiveDict(headers or {})

    payload_args = (data, json, files)
    if payload is None:
        payload = make_payload(*payload_args)
    elif any(i is not None for i in payload_args):
        raise TypeError("`payload` cannot be used with payload arguments: `data`, `json`, `files`")

    return Requisition(verb, url, params=params, headers=headers, payload=payload)




from __future__ import annotations
from typing import Any, Mapping

from functools import cached_property

from ..util.attribute_mapping_proxy import DictAndListRecursiveAttributeMappingProxy


D = Mapping[str, Any]
B = DictAndListRecursiveAttributeMappingProxy[Any]


class DatamementoPropertiesMixin:
    d: D

    @cached_property
    def b(self) -> B:
        return DictAndListRecursiveAttributeMappingProxy(self.d)


class DatamementoBase:
    def __init__(self, d: D) -> None:
        self.d: D = d
        ("")
        self.b: B = DictAndListRecursiveAttributeMappingProxy(d)
        ("")


class DatamementoDataclassesMixin:
    d: D
    b: B

    def __post_init__(self) -> None:
        object.__setattr__(self, 'b', DictAndListRecursiveAttributeMappingProxy(self.d))




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Optional, Callable, Any, Iterable, Mapping
if TYPE_CHECKING:
    from .....client_ASYNC import Client

from ..listing_async_paginator import ListingAsyncPaginator

T = TypeVar('T')

class Time(ListingAsyncPaginator[T]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        limit: Optional[int] = 100,
        params: Optional[Mapping[str, str]] = None,
        cursor_extractor: Callable[[Any], str] = lambda x: x['data']['name'],
    ):
        super().__init__(client, url, limit=limit, params=params, cursor_extractor=cursor_extractor)
        self.time: str = ''
        ("")

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        if self.time:
            yield ('t', self.time)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from .live_thread import LiveThread as BaseLiveThread, LiveUpdate as BaseLiveUpdate

class LiveThread(BaseLiveThread):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")

    async def config(self,
        title: str,
        description: str,
        resources: str,
        nsfw: bool,
    ) -> None:
        await self.client.p.live_thread.config(self.idt, title, description, resources, nsfw)

    async def get_update(self, uuid: str) -> LiveUpdate:
        return await self.client.p.live_thread.get_update(self.idt, uuid)

    async def create_update(self, body: str) -> None:
        await self.client.p.live_thread.create_update(self.idt, body)

    async def strike_update(self, uuid: str) -> None:
        await self.client.p.live_thread.strike_update(self.idt, uuid)

    async def delete_update(self, uuid: str) -> None:
        await self.client.p.live_thread.delete_update(self.idt, uuid)


class LiveUpdate(BaseLiveUpdate):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.message_SYNC import MailboxMessage
    from ..stream_SYNC import IStandardStreamEventSubject
    from ...pagination.paginator import CursorPaginator

from ...models.message_SYNC import ComposedMessage, CommentMessage
from ..stream_SYNC import Stream


def inbox_message_extractor(x: MailboxMessage) -> object:
    if isinstance(x, ComposedMessage):
        return (0, x.id)
    elif isinstance(x, CommentMessage):
        return (1, x.comment.id)
    raise Exception

def get_inbox_message_stream_paginator(client: Client) -> CursorPaginator[MailboxMessage]:
    return client.p.message.pulls.inbox().get_paginator()

def make_inbox_message_stream(paginator: CursorPaginator[MailboxMessage], past: Optional[Iterable[MailboxMessage]] = None, seen: Optional[Iterable[object]] = None) -> IStandardStreamEventSubject[MailboxMessage]:
    return Stream(paginator, inbox_message_extractor, past=past, seen=seen)

def create_inbox_message_stream(client: Client) -> IStandardStreamEventSubject[MailboxMessage]:
    return make_inbox_message_stream(get_inbox_message_stream_paginator(client))


def mentions_message_extractor(x: CommentMessage) -> object:
    return x.comment.id

def get_mentions_message_stream_paginator(client: Client) -> CursorPaginator[CommentMessage]:
    return client.p.message.pulls.mentions().get_paginator()

def make_mentions_message_stream(paginator: CursorPaginator[CommentMessage], past: Optional[Iterable[CommentMessage]] = None, seen: Optional[Iterable[object]] = None) -> IStandardStreamEventSubject[CommentMessage]:
    return Stream(paginator, mentions_message_extractor, past=past, seen=seen)

def create_mentions_message_stream(client: Client) -> IStandardStreamEventSubject[CommentMessage]:
    return make_mentions_message_stream(get_mentions_message_stream_paginator(client))




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .calendar import CalendarWidgetConfiguration

from dataclasses import dataclass

from .widget_ASYNC import Widget

@dataclass(repr=False, eq=False)
class CalendarWidget(Widget):
    google_calendar_id: str
    requires_sync: bool
    configuration: CalendarWidgetConfiguration




from __future__ import annotations
from typing import Mapping

_TRUTHY_WORDS = {'1', 'true', 't', 'on', 'yes', 'y'}
_FALSY_WORDS = {'0', 'false', 'f', 'off', 'no', 'n'}

_BOOLEAN_WORDS: Mapping[str, bool] = {
    **{x: True for x in _TRUTHY_WORDS},
    **{x: False for x in _FALSY_WORDS},
}

def booleanify(s: str) -> bool:
    """Convert a string to boolean.

    An empty string is `False`.

    Raises `ValueError` if the given string is not booleanifiable.

    Booleanifiable strings:

    - Truthy: `{'1', 'true', 't', 'on', 'yes', 'y'}`
    - Falsy: `{'0', 'false', 'f', 'off', 'no', 'n'}`
    """
    if s == '':
        return False
    try:
        return _BOOLEAN_WORDS[s.lower()]
    except KeyError:
        pass
    raise ValueError('Not booleanifiable: %r' % s)




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Optional, Sequence, Iterable, Any
if TYPE_CHECKING:
    from ....client_SYNC import Client

from ...paginator import HasMorePaginator, Bidirectional, CursorPaginator
from ....models.subreddit_user import (
    Moderator,
    ApprovedUser,
    BannedUser,
    MutedUser,
)
from ....model_loaders.subreddit_user import (
    load_moderator,
    load_approved_user,
    load_banned_user,
    load_muted_user,
)

T = TypeVar('T')

class ModerationUsersPaginator(HasMorePaginator[T], Bidirectional, CursorPaginator[T]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        limit: Optional[int] = 100,
    ) -> None:
        super().__init__(limit=limit)
        self.direction: bool = True
        ("")
        self.client: Client = client
        ("")
        self.url: str = url
        ("")
        self._after: str = ''
        self._before: str = ''
        self._has_after: bool = True
        self._has_before: bool = True

    def get_cursor(self) -> str:
        return self._after if self.direction else self._before

    def set_cursor(self, value: str) -> None:
        if self.direction:
            self._after = value
        else:
            self._before = value

    def has_more(self) -> bool:
        return self._has_after if self.direction else self._has_before

    def set_has_more(self, value: bool) -> None:
        if self.direction:
            self._has_after = value
        else:
            self._has_before = value

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        if self.limit is not None:
            yield ('count', str(self.limit))

        if self.direction:
            if self._after:
                yield ('after', self._after)
        else:
            if self._before:
                yield ('before', self._before)

    def _fetch_data(self) -> Any:
        params = dict(self._generate_params())
        root = self.client.request('GET', self.url, params=params)
        after = root['after'] or ''
        before = root['before'] or ''
        self._after = after
        self._before = before
        self._has_after = bool(after)
        self._has_before = bool(before)
        return root


class ModeratorsPaginator(ModerationUsersPaginator[Moderator]):
    def fetch(self) -> Sequence[Moderator]:
        root = self._fetch_data()
        order = root['moderatorIds']
        object_map = root['moderators']
        return [load_moderator(object_map[full_id36]) for full_id36 in order]

class ApprovedUsersPaginator(ModerationUsersPaginator[ApprovedUser]):
    def fetch(self) -> Sequence[ApprovedUser]:
        root = self._fetch_data()
        order = root['approvedSubmitterIds']
        object_map = root['approvedSubmitters']
        return [load_approved_user(object_map[full_id36]) for full_id36 in order]

class BannedUsersPaginator(ModerationUsersPaginator[BannedUser]):
    def fetch(self) -> Sequence[BannedUser]:
        root = self._fetch_data()
        order = root['bannedUserIds']
        object_map = root['bannedUsers']
        return [load_banned_user(object_map[full_id36]) for full_id36 in order]

class MutedUsersPaginator(ModerationUsersPaginator[MutedUser]):
    def fetch(self) -> Sequence[MutedUser]:
        root = self._fetch_data()
        order = root['mutedUserIds']
        object_map = root['mutedUsers']
        return [load_muted_user(object_map[full_id36]) for full_id36 in order]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable, Union, Mapping
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.submission_draft import SubmissionDraft
    from ...models.submission_draft_ASYNC import SubmissionDraftList
    from ...types import JSON_ro

import json

from ...model_loaders.submission_draft import load_public_submission_draft
from ...model_loaders.submission_draft_ASYNC import load_submission_draft_list
from ...util.base_conversion import to_base36

class DraftProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    async def retrieve(self) -> SubmissionDraftList:
        """Retrieve the current user's drafts.

        .. .RETURNS

        :rtype: :class:`~.models.submission_draft_ASYNC.SubmissionDraftList`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        root = await self._client.request('GET', '/api/v1/drafts')
        return load_submission_draft_list(root, self._client)

    async def read_public(self, user: str, uuid: str) -> SubmissionDraft:
        """Read a public draft.

        .. .PARAMETERS

        :param `str` user:
            User name.
        :param `str` uuid:
            Draft UUID.

        .. .RETURNS

        :rtype: `~.models.submission_draft.SubmissionDraft`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `FORBIDDEN`:
               - There is no user context.
               - The specified draft does not exist.
               - You do not have permission to view the draft.

            + `BAD_GATEWAY`:
                The specified ID is not a valid UUID.
            + `NOT_FOUND`:
                The specified draft could not be found.
        """
        url = f"https://gateway.reddit.com/desktopapi/v1/draftpreviewpage/{user}/{uuid}"
        root = await self._client.request('GET', url)
        draft_data = root['drafts'][uuid]
        return load_public_submission_draft(draft_data)

    async def create(self,
        *,
        body: Union[str, Mapping[str, JSON_ro]],
        public: Optional[bool] = None,
        subreddit_id: Optional[int] = None,
        title: Optional[str] = None,
        reply_notifications: Optional[bool] = None,
        spoiler: Optional[bool] = None,
        nsfw: Optional[bool] = None,
        oc: Optional[bool] = None,
        flair_uuid: Optional[str] = None,
        flair_text: Optional[str] = None,
    ) -> str:
        """Create a draft.

        .. .PARAMETERS

        :param body:
        :type body: `Union`\\[`str`, `Mapping`\\[`str`, :class:`~.types.JSON_ro`]]
        :param `Optional[bool]` public:
        :param `Optional[int]` subreddit_id:
        :param `Optional[str]` title:
        :param `Optional[bool]` reply_notifications:
        :param `Optional[bool]` spoiler:
        :param `Optional[bool]` nsfw:
        :param `Optional[bool]` oc:
        :param `Optional[str]` flair_uuid:
        :param `Optional[str]` flair_text:

        .. .RETURNS

        :returns: The UUID of the newly created draft.
        :rtype: `str`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        def g() -> Iterable[tuple[str, str]]:
            if isinstance(body, str):
                yield ('kind', 'markdown')
                yield ('body', body)
            else:
                yield ('kind', 'richtext')
                yield ('body', json.dumps(body))

            if public is not None: yield ('is_public_link', '01'[public])
            if subreddit_id is not None: yield ('subreddit', 't5_' + to_base36(subreddit_id))
            if title is not None: yield ('title', title)
            if reply_notifications is not None: yield ('send_replies', '01'[reply_notifications])
            if spoiler is not None: yield ('spoiler', '01'[spoiler])
            if nsfw is not None: yield ('nsfw', '01'[nsfw])
            if oc is not None: yield ('original_content', '01'[oc])
            if flair_uuid is not None: yield ('flair_id', flair_uuid)
            if flair_text is not None: yield ('flair_text', flair_text)

        root = await self._client.request('POST', '/api/v1/draft', data=dict(g()))
        return root['json']['data']['id']

    async def replace(self,
        uuid: str,
        *,
        body: Union[str, Mapping[str, JSON_ro]],
        public: Optional[bool] = None,
        subreddit_id: Optional[int] = None,
        title: Optional[str] = None,
        reply_notifications: Optional[bool] = None,
        spoiler: Optional[bool] = None,
        nsfw: Optional[bool] = None,
        oc: Optional[bool] = None,
        flair_uuid: Optional[str] = None,
        flair_text: Optional[str] = None,
    ) -> None:
        """Update a draft.

        Every parameter should be specified otherwise their effective default will be used!
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('id', uuid)

            if isinstance(body, str):
                yield ('kind', 'markdown')
                yield ('body', body)
            else:
                yield ('kind', 'richtext')
                yield ('body', json.dumps(body))

            if public is not None: yield ('is_public_link', '01'[public])
            if subreddit_id is not None: yield ('subreddit', 't5_' + to_base36(subreddit_id))
            if title is not None: yield ('title', title)
            if reply_notifications is not None: yield ('send_replies', '01'[reply_notifications])
            if spoiler is not None: yield ('spoiler', '01'[spoiler])
            if nsfw is not None: yield ('nsfw', '01'[nsfw])
            if oc is not None: yield ('original_content', '01'[oc])
            if flair_uuid is not None: yield ('flair_id', flair_uuid)
            if flair_text is not None: yield ('flair_text', flair_text)

        await self._client.request('PUT', '/api/v1/draft', data=dict(g()))

    async def delete(self, uuid: str) -> None:
        """Delete a draft.

        .. .PARAMETERS

        :param `str` uuid:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `VALIDATION_ERRORS`:
               - The specified draft does not exist.
               - The specified draft UUID is not valid.

            + `UNKNOWN_THRIFT_ERROR`:
                The specified draft no longer exists.
        """
        await self._client.request('DELETE', '/api/v1/draft', params={'draft_id': uuid})



pass
'''
from __future__ import annotations
from typing import TYPE_CHECKING, Iterator
if TYPE_CHECKING:
    from .shdlr import Scheduler

from contextvars import ContextVar
from contextlib import contextmanager

_shdlr_ctx_var: ContextVar[Scheduler] = ContextVar('shdlr_ctx')

def get_running_shdlr() -> Scheduler:
    try:
        return _shdlr_ctx_var.get()
    except LookupError:
        raise RuntimeError('no running scheduler') from None

@contextmanager
def enter_shdlr_ctx(shdlr: Scheduler) -> Iterator[None]:
    token = _shdlr_ctx_var.set(shdlr)
    try:
        yield
    finally:
        _shdlr_ctx_var.reset(token)
'''#'''




from . import http_client_SYNC  # noqa: F401
from .http_client_SYNC import HTTPClient as HTTPClient  # noqa: F401
from . import reddit_token_obtainment_client_SYNC  # noqa: F401
from .reddit_token_obtainment_client_SYNC import RedditTokenObtainmentClient as RedditTokenObtainmentClient  # noqa: F401
from . import authorizer_SYNC  # noqa: F401
from .authorizer_SYNC import Authorizer as Authorizer, Authorized as Authorized  # noqa: F401
from . import rate_limited_SYNC  # noqa: F401
from .rate_limited_SYNC import RateLimited as RateLimited  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, IO
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.stylesheet import StylesheetInfo

from ...model_loaders.stylesheet import load_stylesheet_info
from ...http.payload import make_multipart

class SubredditStyleOldProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    def get_stylesheet(self, sr: str) -> StylesheetInfo:
        """Get a subreddit's stylesheet information.

        This endpoint can be called without being a moderator of the target subreddit.
        You don't even have to be logged in.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `private`:
                You do not have access to the specified subreddit; it is private.
            + `banned`:
                You do not have access to the specified subreddit; it is banned.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `302`:
                The subreddit does not exist.
        """
        root = self._client.request('GET', '/about/stylesheet', params={'r': sr})
        return load_stylesheet_info(root['data'])

    def edit_stylesheet(self, sr: str, content: str, *, message: str = '') -> None:
        """Update a subreddit's stylesheet.

        The stylesheet can also be updated by editing the `config/stylesheet` wiki page.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` content:
        :param `str` message:
            A commit message.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `private`:
                You do not have access to the specified subreddit; it is private.
            + `banned`:
                You do not have access to the specified subreddit; it is banned.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The subreddit does not exist.
        """
        data = {'r': sr, 'op': 'save', 'stylesheet_contents': content, 'reason': message}
        self._client.request('POST', '/api/subreddit_stylesheet', data=data)

    def add_stylesheet_image(self, sr: str, name: str, file: IO[bytes]) -> None:
        """Upload an image for use in the subreddit stylesheet.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` name:
            Must be a valid CSS identifier.
        :param `IO[bytes]` file:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `BAD_CSS_NAME`:
                The `name` parameter was not specified or was an invalid CSS identifier.
            + `IMAGE_ERROR`:
                The image file was invalid.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to upload an image to the specified subreddit.
        """
        payload = make_multipart({
            'r': sr,
            'upload_type': 'img',
            'name': name,
            'file': (file, 'file'),
        })
        self._client.request('POST', '/api/upload_sr_img', payload=payload)

    def remove_stylesheet_image(self, sr: str, name: str) -> None:
        """Delete an image from the subreddit's stylesheet custom image set.

        The image will no longer count against the subreddit's image limit,
        however the actual image data may still be accessible for an unspecified
        amount of time. If the image is currently referenced by the subreddit's
        stylesheet, that stylesheet will no longer validate and won't be
        submittable until the image reference is removed.

        If the specified image name does not exist, it is treated as a success.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` name:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `500`:
                The `name` parameter was empty.
        """
        self._client.request('POST', '/api/delete_sr_img', data={'r': sr, 'img_name': name})

    def set_icon(self, sr: str, file: IO[bytes]) -> None:
        """Set the subreddit icon.

        .. .PARAMETERS

        :param `str` sr:
        :param `IO[bytes]` file:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `IMAGE_ERROR`:
                The image file was invalid.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to upload an image to the specified subreddit.
        """
        payload = make_multipart({
            'r': sr,
            'upload_type': 'header',
            'file': (file, 'file'),
        })
        self._client.request('POST', '/api/upload_sr_img', payload=payload)

    def unset_icon(self, sr: str) -> None:
        """Remove the subreddit's icon.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
        """
        self._client.request('POST', '/api/delete_sr_header', data={'r': sr})

    def set_mobile_icon(self, sr: str, file: IO[bytes]) -> None:
        """Set the mobile icon.

        .. .PARAMETERS

        :param `str` sr:
        :param `IO[bytes]` file:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `IMAGE_ERROR`:
                The image file was invalid.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to upload an image to the specified subreddit.
        """
        payload = make_multipart({
            'r': sr,
            'upload_type': 'icon',
            'file': (file, 'file'),
        })
        self._client.request('POST', '/api/upload_sr_img', payload=payload)

    def unset_mobile_icon(self, sr: str) -> None:
        """Remove the mobile icon.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
        """
        self._client.request('POST', '/api/delete_sr_icon', data={'r': sr})

    def set_mobile_banner(self, sr: str, file: IO[bytes]) -> None:
        """Set the mobile banner icon.

        .. .PARAMETERS

        :param `str` sr:
        :param `IO[bytes]` file:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `IMAGE_ERROR`:
                The image file was invalid.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to upload an image to the specified subreddit.
        """
        payload = make_multipart({
            'r': sr,
            'upload_type': 'banner',
            'file': (file, 'file'),
        })
        self._client.request('POST', '/api/upload_sr_img', payload=payload)

    def unset_mobile_banner(self, sr: str) -> None:
        """Remove the mobile banner icon.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
        """
        self._client.request('POST', '/api/delete_sr_banner', data={'r': sr})




from __future__ import annotations

from dataclasses import dataclass

from .widget_ASYNC import Widget

@dataclass(repr=False, eq=False)
class CommunityDetailsWidget(Widget):
    public_description: str
    subscriber_text: str
    viewing_text: str
    subscriber_count: int
    viewing_count: int
